
import datetime
import traceback
import json
import numpy as np
import pandas as pd
import re
from ...api.basic import BasicDataApi
from ...api.raw import RawDataApi
from ..raw.raw_data_helper import RawDataHelper
from .basic_data_helper import BasicDataHelper
from ...manager import DataManager
from ..derived.derived_indicators_processor_group_advanced import FundIndicatorProcessorGroup
from ....constant import IndClassType

class Data:
    
    pass

class FundTagProcessor:

    FUND_CLASSIFIER = {
        'stock': ['普通股票型基金', '偏股混合型基金', '平衡混合型基金', '灵活配置型基金'], 
        'index': ['被动指数型基金', '增强指数型基金'],
    }
    CONCEPT_GROUP_ID = ['dm_cpt_201223','dm_ind_201224']
    THRESHOLD = 0.2

    THRESHOLD_DICT = {
        '军工':0.2,
        '银行':0.2,
        '养老':0.19,
        '白酒':0.2,
        '5G':0.2,
        '芯片':0.2,
        '半导体':0.15,
        '券商':0.2,
        '医疗':0.15,
        '光伏':0.2,
        '新能源':0.2,
        '医美':0.13,
        '黄金':0.15,
        '农业':0.13,
        '食品饮料':0.2,
        '金融':0.2,
        '地产':0.15,
        '有色':0.15,
        '消费':0.2,
        '可选消费':0.18,
        '主要消费':0.2,
        '科技':0.2,
        '医药':0.15,
        '金融地产':0.2,
        }

    INDEX_FUND_RAW = {
        '白酒':['白酒'],
        '食品饮料':['食品'],
        '消费':['消费','可选'],
        '可选消费':['可选'],
        '主要消费':['消费'],
        '5G':['5G'],
        '半导体':['半导体'],
        '科技':['信息','科技','电信'],
        '光伏':['光伏'],
        '农业':['农业'],
        '新能源':['新能源'],
        '医药':['医药'],
        '医疗':['医疗'],
        '银行':['银行'],
        '券商':['非银'],
        '金融地产':['金融地产'],
        '金融':['金融'],
        '地产':['地产'],
        '有色':['有色'],
        '黄金':['黄金'],
        '军工':['军工'],
        '养老':['养老']
        }

    def __init__(self, data_helper:BasicDataHelper):
        self._data_helper = BasicDataHelper()
        self.config = FundIndicatorProcessorGroup
        self.data = Data()

    def data_prepare(self):
        today = datetime.datetime.now().date()
        md = today.strftime('%m%d')
        sentry_date = pd.Series(self.config.REPORT_DATE_LIST)
        sentry = sentry_date[sentry_date <= md]
        # 为半年报/年报发布预留一个季度的缓冲时间
        if sentry.empty:
            real_date = datetime.date(today.year - 1, 6, 30)
        elif len(sentry) <= 2:
            real_date = datetime.date(today.year - 1, 12, 31)
        else:
            real_date = datetime.date(today.year, 6, 30)

        self.data.fund_type_dic = {}
        for fund_type, wind_class_list in self.FUND_CLASSIFIER.items():
            for wind_class in wind_class_list:
                self.data.fund_type_dic[wind_class] = fund_type

        self.data.stock_info = DataManager.raw_data(func_name='get_em_stock_info')
        self.data.stock_info.loc[:,'sw_indus_1'] = self.data.stock_info.bl_sws_ind_code.map(lambda x: x.split('-')[0] if x is not None else None)
        
        self.data.industry_info = DataManager.raw_data(func_name='get_em_industry_info')
        self.data.industry_info = self.data.industry_info[self.data.industry_info.ind_class_type == IndClassType.SWI1]

        self.data.fund_info = DataManager.basic_data(func_name='get_fund_info')
        self.data.fund_info['fund_type'] = self.data.fund_info.wind_class_2.map(lambda x: self.data.fund_type_dic.get(x))
        self.data.stock_fund_list = self.data.fund_info[self.data.fund_info.fund_type == 'stock'].fund_id.tolist()
        self.data.index_fund_list = self.data.fund_info[self.data.fund_info.fund_type == 'index'].fund_id.tolist()
        
        self.data.fund_position = DataManager.basic_data(func_name=f'get_fund_hold_stock_latest', fund_list = self.data.stock_fund_list + self.data.index_fund_list ).set_index('fund_id')
        self.data.concept_data = DataManager.basic_data(func_name=f'get_fund_stock_concept', tag_group_id_list = self.CONCEPT_GROUP_ID)
        self.data.concept_data = self.data.concept_data[['stock_id','tag_name']].drop_duplicates().set_index('stock_id')
        self.fund_position_date_process()
        
        self.data.index_info = DataManager.basic_data(func_name='get_index_info')
        self.data.fund_benchmark = DataManager.basic_data(func_name='get_fund_benchmark')

    def fund_position_date_process(self):
        _res = []
        for r in self.data.fund_position.iterrows():
            fund_id = r[0]
            for i in range(1,11):
                dic = {
                    'fund_id' : fund_id,
                    'stock_id':r[1][f'rank{i}_stock_code'],
                    'stock_weight':r[1][f'rank{i}_stockweight'], 
                }
                _res.append(dic)
        self.data.fund_position = pd.DataFrame(_res).set_index('fund_id')
                
    def get_stocks_concept(self, stock_id):
        if stock_id in self.data.concept_data.index:
            return self.data.concept_data.loc[[stock_id]].tag_name.tolist()
        return []

    def get_fund_holdings(self, fund_id):
        if fund_id in self.data.fund_position.index:
            df = self.data.fund_position.loc[fund_id].copy()
            df.loc[:,'stock_weight'] = df.stock_weight / df.stock_weight.sum()
            df = df.set_index('stock_id').to_dict()['stock_weight']
            return df
        else:
            return None

    def calculate_fund_concept(self, fund_list):
        self.fund_concept_weight_dic = {}
        self.fund_concept_dic = {}
        for fund_id in fund_list:
            self.fund_concept_weight_dic[fund_id] = {}
            fund_weight_dic = self.get_fund_holdings(fund_id)
            if fund_weight_dic is None:
                continue
            for stock_id, stock_weight in fund_weight_dic.items():
                concept_list = self.get_stocks_concept(stock_id)
                for concept_i in concept_list:
                    if concept_i is not self.fund_concept_weight_dic[fund_id]:
                        self.fund_concept_weight_dic[fund_id][concept_i] = stock_weight
                    else:
                        self.fund_concept_weight_dic[fund_id][concept_i] += stock_weight
            self.fund_concept_dic[fund_id] = []
            for concept_i, weight_i in self.fund_concept_weight_dic[fund_id].items():
                threshold = self.THRESHOLD_DICT[concept_i]
                if weight_i >= threshold:
                    self.fund_concept_dic[fund_id].append(concept_i)

    def get_fund_concept_result(self):
        fund_concept = []
        for fund_id, concept_list in self.fund_concept_dic.items():
            if not concept_list == []:
                fund_concept.append({'fund_id':fund_id,'concept':concept_list})
        return pd.merge(pd.DataFrame(fund_concept),self.data.fund_info[['fund_id','desc_name']],on='fund_id')

    def get_fund_concept_weight_result(self):
        res = []
        for fund_id, concept_dic in self.fund_concept_weight_dic.items():
            for concept_i, weight_i in concept_dic.items():
                dic = {
                    'fund_id':fund_id,
                    'concept':concept_i,
                    'weight_i':weight_i,
                }
                res.append(dic) 
        return pd.DataFrame(res)

    def get_concept_id(self, tag_name, raw_concept_id):
        return f'{tag_name}@{str(raw_concept_id)}'

    def calculate_stock_fund_tag(self):
        
        fund_list = self.data.stock_fund_list
        self.calculate_fund_concept(fund_list)
        fund_concept_df = self.get_fund_concept_result()
        fund_concept_weight = self.get_fund_concept_weight_result()
        fund_concept_df = fund_concept_df[['fund_id','concept']].rename(columns={'concept':'tag_name'})
        fund_concept_df = fund_concept_df.explode('tag_name')
        return fund_concept_df

    def select_one_benchmark(self, x):
        if not x:
            return None
        x = json.loads(x)
        select = max(x, key=x.get)
        if x[select] >= 0.8:
            return select
        return None

    def calculate_index_fund_tag(self):
        df = self.data.fund_benchmark[self.data.fund_benchmark.fund_id.isin(self.data.index_fund_list)].copy()
        df.loc[:,'benchmark_id'] = df.benchmark_s_raw.map(lambda x:self.select_one_benchmark(x))
        df = df[['fund_id','index_text','benchmark_id']]
        _df = self.data.index_info[['index_id','desc_name']].rename(columns={'index_id':'benchmark_id'})
        df = pd.merge(df,_df, on='benchmark_id').rename(columns={'desc_name':'index_name'})
        df = pd.merge(df,self.data.fund_info[['fund_id','desc_name']], on='fund_id').rename(columns={'desc_name':'fund_name'})
        index_fund_rules = {}
        for k, vs in self.INDEX_FUND_RAW.items():
            for v in vs:
                if v not in index_fund_rules:
                    index_fund_rules[v] = []
                index_fund_rules[v].append(k)
        result = []
        for tag, tag_detail_list in index_fund_rules.items():
            _res = []
            for tag_i in tag_detail_list:
                _df_i = df[df.index_name.str.contains(tag)].copy()
                if _df_i.shape[0] >= 1:
                    _df_i.loc[:,'tag_name'] = tag_i
                    _res.append(_df_i)
            if _res == []:
                #print(tag)
                continue
            _df = pd.concat(_res).drop_duplicates()
            result.append(_df)
        df = pd.concat(result)[['fund_id','tag_name']]
        return df

    def make_concept_id(self):    
        tag_group_id = 'dm_fund_201224'
        stock_fund_concept_df = self.calculate_stock_fund_tag()
        index_fund_concept_df = self.calculate_index_fund_tag()
        fund_concept_df = pd.concat([stock_fund_concept_df,index_fund_concept_df],axis=0)

        tag_list = fund_concept_df.tag_name.unique().tolist()
        num = 1
        tag_dic = {}
        for tag_id in tag_list:
            tag_dic[tag_id] = self.get_concept_id(tag_group_id, num)
            num += 1
        fund_concept_df.loc[:,'tag_id'] = fund_concept_df.tag_name.map(lambda x: tag_dic[x])
        fund_concept_df.loc[:,'tag_group_id'] = tag_group_id
        name_dic = self.data.fund_info.set_index('fund_id').to_dict()['desc_name']
        name_to_id_dic = self.data.fund_info.set_index('desc_name').to_dict()['fund_id']
        tag_name_list = fund_concept_df.tag_name.unique().tolist()
        # 筛选类型
        result = []
        for tag_i in tag_name_list:
            df_tag_i = fund_concept_df[fund_concept_df.tag_name == tag_i].copy()
            df_tag_i.loc[:,'desc_name'] = df_tag_i.fund_id.map(lambda x : name_dic[x])
            delete_id_list = []
            fund_name_list = df_tag_i.desc_name.tolist()
            for r in df_tag_i.itertuples():
                desc_name = r.desc_name 
                fund_id = r.fund_id
                if 'B' in desc_name or 'C' in desc_name:
                    prefered_fund, count = re.subn(r'[BC]$', 'A', desc_name)
                    if count != 0 and prefered_fund in fund_name_list:
                        _fund_id = name_to_id_dic[prefered_fund]
                        print(f'[BC] {tag_i} remove {prefered_fund} keep {desc_name}')
                        delete_id_list.append(_fund_id)
                if 'H' in desc_name or 'E' in desc_name:
                    prefered_fund, count = re.subn(r'[HE]$', 'A', desc_name)
                    if count != 0 and prefered_fund in fund_name_list:
                        print(f'[EH] {tag_i} remove {desc_name} keep {prefered_fund}')
                        delete_id_list.append(r.fund_id)
            result.append(df_tag_i[~df_tag_i.fund_id.isin(delete_id_list)])
        
        df = pd.concat(result,axis=0)
        # 去掉剩下的 E H R I类
        EHR_PROBLEM_EKYWORDS = {
            'E': ['ETF', 'REITs', 'ESG', 'CES'],
            'H': ['H股', 'A-H50', '50AH', '中证AH'],
            'R': ['REITs'],
            'I': [],
        }
        ehr_funds = []
        for r in df[['desc_name', 'fund_id']].itertuples():
            for type_word, problem_words in EHR_PROBLEM_EKYWORDS.items():
                i = r.desc_name
                for problem_i in problem_words:
                    i = i.replace(problem_i, '')
                if type_word in i:
                    ehr_funds.append(r.fund_id)     
        
        df = df[~df['fund_id'].isin(ehr_funds)]
        self._data_helper._upload_basic(df, 'fund_tag')

    


import numpy as np


# 估值数据
PE_THRESHOLDS = [50, 10]
PB_THRESHOLDS = [8, 1]
DYR_THRESHOLDS = [0.01, 0.05]
PEG_THRESHOLDS = [0.04, 0.16]

# 财务数据
ROE评级阈值 = [0.05, 0.2]
ROIC评级阈值 = [0.03, 0.12]
ROA评级阈值 = [0.005, 0.1]
毛利润率阈值 = [0.05, 0.4]

经营现金流净值比利润总额阈值 = [0.4, 1.2]
所得税比利润总额阈值 = [0.2, 0.02]
营业外支出比利润总额阈值 = [0.005, 0.02]

自由现金流对净利润比评级阈值 = [0.5, 1.25]
资本支出比净利润阈值阈值 = [0.06, 0.03]

流动比率评级阈值 = [0.5, 2.0]
利息保障倍数评级阈值 = [1, 10]

近1年净利润增速阈值 = [0.04, 0.16]
自由现金流年成长率阈值 = [0.04, 0.16]

现金周转天数评级阈值 = [365, 0]


def rating_by_thresholds(df, name, thresholds):
    def rating(value):
        if np.isnan(value):
            return np.NaN

        if value < 0:
            return 1

        if len(thresholds) < 2:
            print('thresholds不能少于2个')

        ascending = thresholds[-1] > thresholds[0]
        if ascending:
            if value <= thresholds[0]:
                return 1
            if value >= thresholds[-1]:
                return 99
            return (value - thresholds[0]) / (thresholds[-1] - thresholds[0]) * 100
        else:
            if value >= thresholds[0]:
                return 1
            if value <= thresholds[-1]:
                return 99
            return (thresholds[0] - value) / (thresholds[0] - thresholds[-1]) * 100

    df[name + '评级'] = df[name].apply(rating)
    df[name + '评级'] = df[name + '评级'].fillna(0)

    return df


def compute_price_ratings(df):

    df = rating_by_thresholds(df, 'TTM.市盈率', PE_THRESHOLDS)
    df = rating_by_thresholds(df, '市净率', PB_THRESHOLDS)
    df = rating_by_thresholds(df, '股息率', DYR_THRESHOLDS)
    df = rating_by_thresholds(df, 'PEG', PEG_THRESHOLDS)
    df['估值评分'] = df['TTM.市盈率评级'] * 0.25 + df['市净率评级'] * 0.25 + df[
        '股息率评级'] * 0.25 + df['PEG评级'] * 0.25

    return df


def compute_finance_ratings(df):
    # 盈利能力
    df = rating_by_thresholds(df, 'TTM.平均ROE', ROE评级阈值)
    df = rating_by_thresholds(df, 'TTM.ROIC', ROIC评级阈值)
    df = rating_by_thresholds(df, 'TTM.ROA', ROA评级阈值)
    df = rating_by_thresholds(df, 'TTM.毛利润率', 毛利润率阈值)
    df['盈利能力评级'] = df['TTM.平均ROE评级'] * 0.25 + df['TTM.ROIC评级'] * 0.25 + df['TTM.ROA评级'] * 0.25 + df['TTM.毛利润率评级'] * 0.25
    # 收益质量
    df = rating_by_thresholds(df, 'TTM.经营现金流净值比利润总额', 经营现金流净值比利润总额阈值)
    df = rating_by_thresholds(df, 'TTM.所得税比利润总额', 所得税比利润总额阈值)
    df = rating_by_thresholds(df, 'TTM.营业外支出比利润总额', 营业外支出比利润总额阈值)
    df['收益质量评级'] = df['TTM.经营现金流净值比利润总额评级'] * 0.4 + df['TTM.所得税比利润总额评级'] * 0.3 + df['TTM.营业外支出比利润总额评级'] * 0.3

    # 现金流量
    df = rating_by_thresholds(df, 'TTM.自由现金流对净利润比', 自由现金流对净利润比评级阈值)
    df = rating_by_thresholds(df, 'TTM.资本支出比净利润', 资本支出比净利润阈值阈值)
    df['现金流量评级'] = df['TTM.自由现金流对净利润比评级'] * 0.6 + df['TTM.资本支出比净利润评级'] * 0.4

    # 运营效率
    df = rating_by_thresholds(df, '现金周转天数', 现金周转天数评级阈值)
    df['运营效率评级'] = df['现金周转天数评级']

    # 资本结构
    df = rating_by_thresholds(df, '流动比率', 流动比率评级阈值)
    df = rating_by_thresholds(df, '利息保障倍数', 利息保障倍数评级阈值)
    df['资本结构评级'] = df['流动比率评级'] * 0.5 + df['利息保障倍数评级'] * 0.5

    # 成长能力
    df = rating_by_thresholds(df, '单季度.自由现金流年成长率', 自由现金流年成长率阈值)
    df = rating_by_thresholds(df, '近1年净利润增速', 近1年净利润增速阈值)
    df['成长能力评级'] = df['单季度.自由现金流年成长率评级'] * 0.5 + df['近1年净利润增速评级'] * 0.5

    df['财务评分'] = df['盈利能力评级'] * 0.2 + df['收益质量评级'] * 0.2 + df[
        '现金流量评级'] * 0.2 + df['运营效率评级'] * 0.2 + df['资本结构评级'] * 0.2 + df['成长能力评级'] * 0.2

    return df





import traceback
from typing import List, Dict
from functools import partial
from datetime import datetime

import pandas as pd
import statsmodels.api as sm
from sklearn.preprocessing import StandardScaler

from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...api.derived import DerivedDataApi
from ...view.basic_models import StyleAnalysisStockFactor
from ...view.derived_models import StyleAnalysisFactorReturn

from .derived_data_helper import DerivedDataHelper


class StyleAnalysisProcessor:

    # 一些预定义变量, 永远不应该修改
    _STOCK_FACTOR_RETURN_SOURCE_INDEX = 'all'
    _WINDOW_LEN = 60

    def __init__(self, data_helper, universe: str = _STOCK_FACTOR_RETURN_SOURCE_INDEX):
        self._data_helper = data_helper
        self._universe: str = universe

        # 获取除const因子(市场因子)以外, basic的表中所有因子列的名字
        self._factor_columns: List[str] = StyleAnalysisStockFactor.__table__.columns.keys()[:]
        try:
            for c in ('stock_id', 'datetime', 'rate_of_return', '_update_time'):
                self._factor_columns.remove(c)
        except ValueError as e:
            print(f"it's weird, got a error {e} while dropping some columns that is not a factor")

        # universe中股票成分列表
        self._stocks: pd.DataFrame = None
        # 因子收益
        self._factor_return: List[pd.Series] = []
        # 横截面回归的r2
        self._r2: List[float] = []
        # 横截面回归时每一天的样本容量
        self._sample_nums: List[float] = []
        # 横截面回归时每一天universe中实际的样本容量
        self._real_sample_nums: List[float] = []

        # 基金收益归因信息
        self._fund_attribution: Dict[str, pd.DataFrame] = {}
        self._fund_attribution_list: List[pd.Series] = []

        # 因子暴露(时间序列回归系数)信息
        self._exposure_estimated: Dict[str, pd.DataFrame] = {}
        self._exposure_estimated_list: List[pd.Series] = []

    @property
    def STOCK_FACTOR_RETURN_SOURCE_INDEX(self):
        return self._STOCK_FACTOR_RETURN_SOURCE_INDEX

    @property
    def FACTOR_COLUMNS(self):
        return self._factor_columns

    @property
    def WINDOW_LEN(self):
        return self._WINDOW_LEN

    def _section_regression(self, x: pd.DataFrame):
        # x是截面上(一个时间点上)universe中所有股票的信息
        df = x.dropna().reset_index(level='stock_id')
        if df.shape[0] == 0:
            # 如果这一天没有universe中任何股票的信息，则直接返回(无法做回归)
            return

        # 记一下样本容量
        self._sample_nums.append(df.shape[0])

        # 根据这一天的日期获取对应时间点上的指数成分信息
        assert df.index.nunique() == 1, 'the number of unique values of index in section regression must be 1!!'
        if self._stocks is not None:
            stock_list = self._stocks.loc[df.index.date[0], 'stock_list'].split(',')
            # 只使用成分股做回归
            df = df[df['stock_id'].isin(stock_list)]

        # 记一下根据指数成分过滤后实际的样本容量
        self._real_sample_nums.append(df.shape[0])

        # 对每种风格因子做标准化
        normalized = StandardScaler().fit_transform(df.loc[:, self.FACTOR_COLUMNS].to_numpy())
        # 收益率对const(市场因子)+所有风格因子做横截面回归
        results = sm.OLS(df.loc[:, 'rate_of_return'].to_numpy(), sm.add_constant(normalized)).fit()
        # 记录r2以及rate_of_return
        self._r2.append(results.rsquared)
        # 将系数(因子收益率)保存下来, 然后再加上日期和universe
        return pd.Series(dict(zip(['const']+self.FACTOR_COLUMNS, results.params.tolist()))).append(pd.Series({'universe_index': self._universe}))

    def calculate(self, start_date: str, end_date: str):
        start_date = datetime.strptime(start_date, '%Y%m%d').date().isoformat()
        end_date = datetime.strptime(end_date, '%Y%m%d').date().isoformat()

        # 由于需要shift, 所以需要从start_date往前多取一些数据
        trading_day = BasicDataApi().get_trading_day_list().set_index('datetime')
        start_date_loc = trading_day.index.get_loc(datetime.strptime(start_date, '%Y-%m-%d').date(), method='ffill')
        assert isinstance(start_date_loc, int), 'start date is not a valid trading day'

        extra_days = 2
        assert start_date_loc >= extra_days, 'there is no enough extra days before start date'

        start_date_with_extra_days = trading_day.index.array[start_date_loc-extra_days]
        print(f'start date with extra days: {start_date_with_extra_days}')

        # 获取指数成分信息(时间序列)
        if self._universe != self.STOCK_FACTOR_RETURN_SOURCE_INDEX:
            self._stocks = RawDataApi().get_em_index_component(start_date_with_extra_days, end_date, [self._universe]).set_index('datetime')
            print('load index component data done')
        else:
            print('calculation on whole market, no need to get index component')

        # 获取风格分析因子值
        factor_values = BasicDataApi().get_style_analysis_data(start_date_with_extra_days, end_date).drop(columns='_update_time')
        print('load style analysis data done')

        # 收益率列, 存在一些日期的股票没有值, 如未上市等, 这里不做处理后边直接dropna
        rate_of_return: pd.DataFrame = factor_values.pivot_table(index='datetime', columns='stock_id', values='rate_of_return')
        if self._stocks is not None:
            # 指数成分信息与因子值进行索引对齐(空值bfill填充, 因为没有前值)
            self._stocks = self._stocks.reindex(rate_of_return.index, method='bfill')
            assert not self._stocks.isnull().to_numpy().any(), 'index component df should not have any NaN!!'
        # 收益率列向上移1行, 使得其他列T-1日对应T日收益率
        factor_values = factor_values.set_index(['datetime', 'stock_id'])
        factor_values['rate_of_return'] = rate_of_return.shift(-1).stack()

        # 对每个时间点做横截面回归
        self._factor_return = factor_values.groupby(level='datetime').apply(self._section_regression)
        # 最后再整体下移1行, 使得日期与收益率的日期对齐(均为T日, 之前为T-1日对应T日收益率)
        self._factor_return = self._factor_return.shift(1).dropna()

    def process(self, start_date, end_date):
        failed_tasks = []
        try:
            self.calculate(start_date, end_date)
            # 将风格因子收益率的时间序列存到DB
            self._data_helper._upload_derived(self._factor_return.loc[start_date:end_date, :].reset_index(), StyleAnalysisFactorReturn.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append(f'style_analysis_factor_return({self._universe})')
        return failed_tasks

    # def analyse_stock(self, start_date: str, end_date: str, stock_ids: List[str]):
    #     # 取一只个股的收益率时间序列
    #     stock_price = RawDataApi().get_em_stock_price(start_date, end_date, stock_ids, ['stock_id', 'datetime', 'close', 'pre_close']).set_index('datetime')
    #     df = stock_price.pivot_table(index='datetime', columns='stock_id', values='close') / stock_price.pivot_table(index='datetime', columns='stock_id', values='pre_close') - 1
    #     print(df)

    #     # style_factor_return = DerivedDataApi().get_style_factor_return(start_date, end_date, self.STOCK_FACTOR_RETURN_SOURCE_INDEX).set_index('datetime')
    #     style_factor_return = pd.DataFrame(self._factor_return)
    #     for stock in df.columns:
    #         # 个股的收益率时间序列对风格因子收益率的时间序列做回归
    #         results = sm.OLS(df.loc[:, stock].to_numpy(), style_factor_return.loc[:, self.FACTOR_COLUMNS].to_numpy()).fit()
    #         return pd.Series(dict(zip(self.FACTOR_COLUMNS, results.params.tolist()))).append(pd.Series({'universe_index': self.STOCK_FACTOR_RETURN_SOURCE_INDEX}))
    #         # print(df.loc[:, stock] - style_factor_return.loc[:, self.FACTOR_COLUMNS] * regr.coef_)

    def _lambda_2(self, x, whole):
        # x是基金每日收益率, 对相应日期的const+风格因子收益做回归(加前置的常数项)(使用前WINDOW_LEN日)
        results = sm.OLS(x[:-1].to_numpy(), sm.add_constant(whole.loc[x.index[:-1], ['const']+self.FACTOR_COLUMNS].to_numpy())).fit()
        # 回归结果中包含常数项系数，这里只取另外几个indicator的系数(即上边的const+风格因子收益)
        # T+1日各因子收益的贡献 = T+1日因子收益 * 前WINDOW_LEN日回归系数
        contribution = whole.loc[x.index[-1], ['const']+self.FACTOR_COLUMNS] * results.params[1:]
        # T+1日特异收益率 = T+1日基金收益 - T+1日各因子收益的贡献和
        spec_ret = x.array[-1] - sum(contribution)

        # 记录T+1日各因子收益的贡献, 特异收益率, 基金收益率等信息
        self._fund_attribution_list.append(contribution.append(pd.Series({'spec_ret': spec_ret, 'rate_of_return': x.array[-1], 'datetime': x.index.array[-1]})))
        # 记录对前WINDOW_LEN日回归得到的因子暴露(回归系数), r2等信息
        self._exposure_estimated_list.append(pd.Series(dict(zip(['const']+self.FACTOR_COLUMNS, results.params[1:].tolist()))).append(pd.Series({'r2': results.rsquared, 'datetime': x.index.array[-1]})))
        # 这里简单起见未使用返回值, 而是直接将信息记录在类中
        return 0

    def _lambda_1(self, style_factor_return, fund_info, x):
        # 获取基金的universe
        index_id = (fund_info[fund_info['fund_id'] == x.name].loc[:, 'index_id'].array[0])
        # 取出所有该universe下的所有风格因子收益
        style_factor_return = style_factor_return[style_factor_return['universe_index'] == index_id].set_index('datetime')
        # 将基金的复权净值列加到风格因子收益df里做成一个整体
        # 先统一index, 且基金每日收益率的nan值用0填充
        # 然后dropna, 这样后边rolling的窗口数可以保证是真实的
        whole = style_factor_return.assign(rate_of_return=x.reindex(style_factor_return.index)).dropna()
        # 对这张完整df做rolling
        whole.loc[:, 'rate_of_return'].rolling(window=self.WINDOW_LEN+1).apply(func=self._lambda_2, kwargs={'whole': whole})

        self._fund_attribution[x.name] = pd.DataFrame(self._fund_attribution_list)
        self._exposure_estimated[x.name] = pd.DataFrame(self._exposure_estimated_list)
        # print(self._fund_attribution[x.name])
        print(self._fund_attribution[x.name]['spec_ret'])

    def analyse_fund(self, start_date: str, end_date: str, fund_ids: List[str]):
        # 取基金的收益率时间序列
        fund_nav = BasicDataApi().get_fund_nav(fund_ids)
        # 复权净值的nan使用上一日的值
        df = fund_nav.pivot_table(index='datetime', columns='fund_id', values='adjusted_net_value').fillna(method='ffill')
        assert not df.isnull().to_numpy().any(), 'fund nav df should not have any NaN!!'
        # 计算每日净值增长率
        df = df / df.shift(1) - 1

        # 从fund_info中获取fund_id和fund_index
        fund_info = BasicDataApi().get_fund_info().loc[:, ['fund_id', 'index_id']]
        # 只需要我们要分析的基金的信息
        fund_info = fund_info[fund_info['fund_id'].isin(fund_ids)]
        # 这里还是先用传进来的 universe 来做 index 了
        fund_info['index_id'] = self._universe
        # 将基金信息中的index_id作为universe获取风格因子收益
        style_factor_return = DerivedDataApi().get_style_factor_return(start_date, end_date, fund_info['index_id'].to_list())
        # TODO do a reindex to stock by fund index
        # style_factor_return.reindex(fund_nav.index)

        # 对每一列(即每只基金)进行计算, apply进去的是该基金的复权净值时间序列
        df.apply(partial(self._lambda_1, style_factor_return, fund_info))


if __name__ == "__main__":
    data_helper = DerivedDataHelper()
    for universe in ('hs300', 'csi800', 'all'):
        sap = StyleAnalysisProcessor(data_helper, universe)
        sap.process('20210601', '20210601')



from typing import Optional, Dict
import statsmodels.api as sm
import numpy as np
import pandas as pd
import scipy.stats as st

from ...constant import StockFactorType
from .base import Factor
from .basic_factors import StockPostPriceFactor, StockTradableFactor, StockOpenPriceFactor
from .utils import get_quarterly_tradingday_list


def MAD(x):
    y = (x - x.median()).apply(np.abs)
    ub = x.median() + 5 * y.median()
    lb = x.median() - 5 * y.median()
    x[x > ub] = ub
    x[x < lb] = lb
    return x


def norm(x):
    return (x - x.mean()) / x.std()


class DerivedFactor(Factor):

    _registry = {}
    _factor_names_checker: Dict[str, Factor] = {}

    def __init__(self, name, f_type: StockFactorType):
        super().__init__(name, f_type)
        # 这里我们将每一个universe下的因子值都存下来
        self._factor_ret: Dict[str, pd.DataFrame] = {}
        self._deps.add(StockPostPriceFactor())
        self._deps.add(StockTradableFactor())
        self._deps.add(StockOpenPriceFactor())

    def get_ret(self, universe: str = 'default') -> Optional[pd.DataFrame]:
        try:
            return self._factor_ret[universe]
        except KeyError:
            return self._load_ret(universe)

    def save_ret(self, universe: str = 'default') -> bool:
        try:
            self._factor_ret[universe].to_parquet(self._get_s3_factor_ret_uri(universe), compression='gzip')
            return True
        except KeyError:
            return False

    def register(self):
        def _check_factor_name(f: Factor):
            for one in f._deps:
                _check_factor_name(one)
            try:
                assert id(DerivedFactor._factor_names_checker[f.name]) == id(f), f'duplicated factor with name {f.name}!!!'
            except KeyError:
                DerivedFactor._factor_names_checker[f.name] = f

        DerivedFactor._registry[self.name] = self
        # 检查是否有重名的(名字写错的)factor
        _check_factor_name(self)

    def calc_ret(self, universe: str = 'default', n=5) -> pd.DataFrame:
        '''
        Name: calc_ret
        Input: A DataFrame of factor; number of groups.
        Output: The factor return series of F1-Fn portfolio.
        '''
        ret_list = []
        dt_list = []
        isNaN = True
        factor = self.get(universe=universe)
        # 计算收益
        spp_factor = StockPostPriceFactor().get()
        ret = spp_factor.apply(np.log).diff()
        factor = factor.reindex(columns=ret.columns)
        tradingday_list = get_quarterly_tradingday_list()
        for i in range(0, len(factor)):
            dt_list.append(factor.index[i])
            if len(factor.iloc[i].dropna()) == 0:
                ret_list.append(np.nan)
            elif isNaN:
                F1_bound = factor.iloc[i].dropna().quantile(1 - 1 / n)
                Fn_bound = factor.iloc[i].dropna().quantile(1 / n)
                F1_set = factor.iloc[i][factor.iloc[i] > F1_bound].index
                Fn_set = factor.iloc[i][factor.iloc[i] < Fn_bound].index
                ret_list.append(np.nan)
                isNaN = False
            elif factor.index[i] in tradingday_list:
                F1_ret = ret.iloc[i][F1_set].mean()
                Fn_ret = ret.iloc[i][Fn_set].mean()
                ret_list.append(F1_ret - Fn_ret)

                F1_bound = factor.iloc[i - 1].dropna().quantile(1 - 1 / n)
                Fn_bound = factor.iloc[i - 1].dropna().quantile(1 / n)
                F1_set = factor.iloc[i - 1][factor.iloc[i - 1] >= F1_bound].index
                Fn_set = factor.iloc[i - 1][factor.iloc[i - 1] <= Fn_bound].index
            else:
                F1_ret = ret.iloc[i][F1_set].mean()
                Fn_ret = ret.iloc[i][Fn_set].mean()
                ret_list.append(F1_ret - Fn_ret)

        self._factor_ret[universe] = pd.DataFrame(ret_list, index=dt_list)
        self._factor_ret[universe].columns = [self.name]

    def ic_analysis(self, lis=[1, 5, 10]):
        ic_total = []
        pv_total = []
        for num in lis:
            ic_list = []
            pv_list = []
            factor = self.get()
            ret_table = StockPostPriceFactor().get().apply(np.log).diff()
            length = len(ret_table)
            for i in range(num, length):
                temp = pd.concat([factor.iloc[i - num], ret_table.iloc[i]], axis=1).dropna()
                x = MAD(temp.iloc[:, 0])
                y = MAD(temp.iloc[:, 1])
                ic_list.append(st.spearmanr(x, y).correlation)
                pv_list.append(st.spearmanr(x, y).pvalue)
            df = pd.DataFrame([ic_list, pv_list])
            df = df.T
            df.columns = ['ic', 'pvalue']
            ic_array = df.set_index(ret_table.index[num:])
            ic_total.append(ic_array.ic)
            pv_total.append(ic_array.pvalue)
        ic_data = pd.concat(ic_total, axis=1)
        self.ic_array = pd.concat(ic_total, axis=1)
        self.ic_array.columns = lis
        ic_summary_table = pd.DataFrame()
        ic_summary_table["IC Mean"] = ic_data.mean()
        ic_summary_table["IC Std."] = ic_data.std()
        ic_summary_table["Risk-Adjusted IC"] = ic_data.mean() / ic_data.std()
        t_stat, p_value = st.ttest_1samp(ic_data, 0, nan_policy='omit')
        ic_summary_table["t-stat(IC)"] = t_stat
        ic_summary_table["p-value(IC)"] = p_value
        ic_summary_table["IC Skew"] = st.skew(ic_data, nan_policy='omit')
        ic_summary_table["IC Kurtosis"] = st.kurtosis(ic_data, nan_policy='omit')
        # ic_summary_table.columns = lis
        self.ic_summary = ic_summary_table
        self.ic_summary.index = lis

    def ret_analysis(self, lis=[1, 5, 10]):
        ret = StockPostPriceFactor().get().apply(np.log).diff()
        factor = self.get()
        ret_total = []
        for num in lis:
            b_list = []
            for i in range(len(ret) - num):
                X = sm.add_constant(factor.iloc[i].dropna())
                if len(X) == 0:
                    b_list.append([np.nan, np.nan])
                    continue
                Y = (ret.iloc[i + num].dropna())
                X.columns = ['const', 'factor']
                X.factor = norm(MAD(X.factor))
                Y = norm(MAD(Y))
                idx = (set(Y.index.values) & set(X.index.values))
                model = sm.WLS(Y.loc[idx], X.loc[idx])
                resu = model.fit()
                b_list.append(resu.params)
            qwq = pd.DataFrame(b_list)
            qwq.columns = ['const', 'factor']
            qwq = qwq.set_index(ret.index[num:])
            ret_total.append(qwq.factor)
        ret_data = pd.concat(ret_total, axis=1)
        self.ret_array = ret_data
        self.ret_array.columns = lis
        ret_summary_table = pd.DataFrame()
        ret_summary_table["Annualized Return"] = ret_data.mean() * 240
        ret_summary_table["Annualized Vol"] = ret_data.std() * np.sqrt(240)
        ret_summary_table["Sharpe Ratio"] = ret_data.mean() / ret_data.std()
        t_stat, p_value = st.ttest_1samp(ret_data, 0, nan_policy='omit')
        ret_summary_table["t-stat(Daily Return)"] = t_stat
        ret_summary_table["p-value(Daily Return)"] = p_value
        ret_summary_table["Return Skew"] = st.skew(ret_data, nan_policy='omit')
        ret_summary_table["Return Kurtosis"] = st.kurtosis(ret_data, nan_policy='omit')
        self.ret_summary = ret_summary_table
        self.ret_summary.index = lis

    def N_fold_analysis(self, n=5):
        megedata = pd.DataFrame()
        #megedata["period"]=StockPostPriceFactor().get().pct_change(1).shift(-1).stack()#以收盘价交易
        megedata["period"] = StockOpenPriceFactor().get().pct_change().shift(-2).stack()  #以开盘价交易
        test_factor = self.get()
        tradable = StockTradableFactor().get()
        test_factor = test_factor.replace([np.inf, -np.inf], np.nan)
        clean_factor_data = megedata
        input_factor = test_factor * tradable
        input_factor = input_factor.stack()
        clean_factor_data["factor"] = input_factor
        clean_factor_data = clean_factor_data.dropna()
        clean_factor_data["factor_quantile"] = clean_factor_data["factor"].groupby(level=0).apply(lambda x: ((pd.qcut(x.rank(), n, labels=False, duplicates='drop') + 1)))
        self.clean_factor_data = clean_factor_data

    def _load_ret(self, universe='default') -> Optional[pd.DataFrame]:
        try:
            self._factor_ret[universe]: pd.DataFrame = pd.read_parquet(self._get_s3_factor_ret_uri(universe))
        except Exception as e:
            print(f'retrieve ret from s3 failed, (err_msg){e}; try to re-calc')
            self.calc_ret(universe)
        return self._factor_ret[universe]



import gzip
import json
import boto3
import pandas as pd
import numpy as np
from surfing.data.wrapper.mysql import DerivedDatabaseConnector
from surfing.data.api.derived import DerivedDataApi
from surfing.data.view.derived_models import AllocationDistribution
s3 = boto3.resource('s3')

def get_effective_frontier(bucket:str='effective-frontier-result',key:str='effective_frontier_bk_v0'):
    obj = s3.Object(bucket, key)
    body = obj.get()['Body'].read()
    data = gzip.decompress(body).decode('utf-8')
    data = json.loads(data)
    data = pd.DataFrame(data)
    return data

def upload_derived(df, table_name):
    print(table_name)
    df.to_sql(table_name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')

def build_allocation_distribution():
    bk_result = get_effective_frontier()
    annual_ret_max = bk_result.annual_ret.max()
    mdd_max = bk_result[bk_result.annual_ret == annual_ret_max].mdd.values[0]
    bk_result = bk_result[bk_result.mdd <= mdd_max]
    mdd_min = bk_result.mdd.min()
    bar = int((mdd_max - mdd_min) / 0.0025 )
    res = []
    mdd_steps = list(np.linspace(mdd_min, mdd_max, bar))
    for mdd_b in mdd_steps[:-1]:
        mdd_e = mdd_steps[mdd_steps.index(mdd_b) + 1]
        idx = bk_result[(bk_result.mdd >= mdd_b) & (bk_result.mdd <= mdd_e) ].annual_ret.idxmax()
        dic = bk_result.loc[idx].to_dict()
        dic['mdd_up_limit'] = mdd_e
        res.append(dic)
    df = pd.DataFrame(res)
    df['cash'] = 1 - df.csi500-df.gem-df.gold-df.hs300-df.mmf-df.national_debt-df.sp500rmb
    df['version'] = 1
    upload_derived(df, AllocationDistribution.__table__.name)

def get_saa_from_mdd(mdd:float=0.15):
    df = DerivedDataApi().get_allocation_distribution()
    df = df[df.mdd < mdd]
    if df.empty:
        return {}
    else:
        return df.sort_values('mdd', ascending=False).iloc[0].to_dict()

if __name__ == "__main__":
    # normal return 
    get_saa_from_mdd(mdd=0.20)

    # abnormal return 
    get_saa_from_mdd(mdd=0.01)



from typing import Dict

import numpy as np
import pandas as pd

from .factor_types import ValueFactor, MixFactorBase
from .basic_factors import MarketValueFactor, AdjNetAssetAvgFactor, NetProfitTTMFactor, TotalRevenueTTMFactor
from .basic_factors import EBITDATTMFactor, EntValueFactor, DividendYearlyFactor
from .utils import normalize


class NAToMVFactor(ValueFactor):
    # 净市率
    def __init__(self):
        super().__init__('na_mv')
        self._deps.add(AdjNetAssetAvgFactor())
        self._deps.add(MarketValueFactor())

    def calc(self):
        # 净资产 / 总市值，即市净率倒数
        self._factor = AdjNetAssetAvgFactor().get() / MarketValueFactor().get().replace({0: np.nan})


class NPToMVFactor(ValueFactor):
    # 盈利收益比
    def __init__(self):
        super().__init__('np_mv')
        self._deps.add(NetProfitTTMFactor())
        self._deps.add(MarketValueFactor())

    def calc(self):
        # 净利润 / 总市值，即市盈率倒数
        self._factor = NetProfitTTMFactor().get() / MarketValueFactor().get().replace({0: np.nan})


class SPSToPFactor(ValueFactor):
    # 营收股价比
    def __init__(self):
        super().__init__('sps_p')
        self._deps.add(TotalRevenueTTMFactor())
        self._deps.add(MarketValueFactor())

    def calc(self):
        # 每股销售额 / 股价，即市销率倒数
        self._factor = TotalRevenueTTMFactor().get() / MarketValueFactor().get().replace({0: np.nan})


class DividendYieldFactor(ValueFactor):
    # 股息率
    def __init__(self):
        super().__init__('dys')
        self._deps.add(DividendYearlyFactor())
        self._deps.add(MarketValueFactor())

    def calc(self):
        # 过去12个月总派息额 / 总市值
        self._factor = DividendYearlyFactor().get() / MarketValueFactor().get().replace({0: np.nan})


class EBITDAToMVFactor(ValueFactor):
    # EBITDA股价比
    def __init__(self):
        super().__init__('eb_mv')
        self._deps.add(EBITDATTMFactor())
        self._deps.add(MarketValueFactor())

    def calc(self):
        # EBITDA（税息折旧及摊销前利润） / 总市值
        self._factor = EBITDATTMFactor().get() / MarketValueFactor().get().replace({0: np.nan})


class EBITDAToEVFactor(ValueFactor):
    # EBITDA企业价值比
    def __init__(self):
        super().__init__('eb_ev')
        self._deps.add(EBITDATTMFactor())
        self._deps.add(EntValueFactor())

    def calc(self):
        # EBITDA（税息折旧及摊销前利润） / 企业价值, 即企业倍数的倒数
        self._factor = EBITDATTMFactor().get() / EntValueFactor().get().replace({0: np.nan})


class MixValueFactor(MixFactorBase, ValueFactor):
    # 合成因子
    def __init__(self):
        super(MixFactorBase, self).__init__('mixv')
        self._factor: Dict[pd.DataFrame] = {}
        self._deps.add(NAToMVFactor())
        self._deps.add(NPToMVFactor())
        self._deps.add(SPSToPFactor())
        self._deps.add(DividendYieldFactor())
        self._deps.add(EBITDAToMVFactor())
        self._deps.add(EBITDAToEVFactor())

    def calc(self, universe: str):
        index = NAToMVFactor().get(universe).index
        columns = NAToMVFactor().get(universe).columns
        self._factor[universe] = pd.DataFrame((normalize(NAToMVFactor().get(universe).T) + normalize(NPToMVFactor().get(universe).T) + normalize(SPSToPFactor().get(universe).T) +
                                               normalize(DividendYieldFactor().get(universe).T) + normalize(EBITDAToMVFactor().get(universe).T) + normalize(EBITDAToEVFactor().get(universe).T)).T, index=index, columns=columns)



import time
import json

from ...constant import FOFTradeStatus, HoldingAssetType
from .raw import *
from .basic import *
from .derived import *
from ..view.TL_models import *
from ..wrapper.mysql import TLDatabaseConnector, TestDatabaseConnector
from ..view.test_models import *
class ResearchApi(metaclass=Singleton):
    
    def get_index_quote(self,start_date,end_date, index_ids):
        """指数行情查询接口"""
        try:
            with BasicDatabaseConnector().managed_session() as db_session:
                rename_dic = {
                    'total_turnover': 'amount',
                }
                query = db_session.query(
                        IndexPrice.index_id,
                        IndexPrice.datetime,
                        IndexPrice.open,
                        IndexPrice.high,
                        IndexPrice.low,
                        IndexPrice.close,
                        IndexPrice.volume,
                        IndexPrice.total_turnover
                        
                    ).filter(
                        IndexPrice.index_id.in_(index_ids),
                        IndexPrice.datetime >= start_date, #根据日期筛选数据
                        IndexPrice.datetime <= end_date
                    )           
                df1 = pd.read_sql(query.statement, query.session.bind).rename(columns=rename_dic)
                
            with RawDatabaseConnector().managed_session() as db_session:
                rename_dic = {
                    'index_date':'datetime',
                }
                query = db_session.query(
                        HFIndexPrice.index_id,
                        HFIndexPrice.index_date,
                        HFIndexPrice.close
                    ).filter(
                        HFIndexPrice.index_id.in_(index_ids),
                        HFIndexPrice.index_date >= start_date, #根据日期筛选数据
                        HFIndexPrice.index_date <= end_date
                    )
                df2 = pd.read_sql(query.statement, query.session.bind).rename(columns=rename_dic) 
                
            with TestDatabaseConnector().managed_session() as db_session:
                query = db_session.query(
                        PfIndexPrice.index_id,
                        PfIndexPrice.datetime,
                        PfIndexPrice.close
                    ).filter(
                        PfIndexPrice.index_id.in_(index_ids),
                        PfIndexPrice.datetime >= start_date, #根据日期筛选数据
                        PfIndexPrice.datetime <= end_date
                    )
                df3 = pd.read_sql(query.statement, query.session.bind).rename(columns=rename_dic) 
            df = pd.concat([df1, df2, df3])
            return df
        except Exception as e:
            print('Failed to get data <err_msg> {}'.format(e) + ' from BasicDataApi.get_index_price')
            return None
            
    def get_index_info(self, desc_name:list=[]):
        '''指数基本信息模糊查询接口'''
        try:
            with BasicDatabaseConnector().managed_session() as db_session:
                query = db_session.query(
                            IndexInfo.index_id,
                            IndexInfo.desc_name,
                            IndexInfo.em_id,
                        )
                for name_i in desc_name:
                    query = query.filter(
                            IndexInfo.desc_name.like(f'%{name_i}%')
                        )
                df1 = pd.read_sql(query.statement, query.session.bind)
                dic = {
                    'real_id':'index_id',
                    'real_name':'desc_name',
                    'asset_id':'em_id',
                }
                query = db_session.query(
                            AssetInfo.asset_id,
                            AssetInfo.real_name,
                            AssetInfo.real_id,
                )
                for name_i in desc_name:
                    query = query.filter(
                            AssetInfo.real_name.like(f'%{name_i}%')
                        )
                df2 = pd.read_sql(query.statement, query.session.bind)
                df2 = df2.rename(columns=dic)[dic.values()]
                
            with TestDatabaseConnector().managed_session() as db_session:
                query = db_session.query(
                            PfIndexInfo.index_id,
                            PfIndexInfo.desc_name,       
                )
                for name_i in desc_name:
                    query = query.filter(
                            PfIndexInfo.desc_name.like(f'%{name_i}%')
                        )
                df3 = pd.read_sql(query.statement, query.session.bind)
                df = pd.concat([df1,df2,df3])
                return df
                
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from {ResearchApi.get_index_info}')

    def get_index_info_by_id(self, index_list:list=[]):
        '''指数基本信息查询接口'''
        try:
            with BasicDatabaseConnector().managed_session() as db_session:
                query = db_session.query(
                            IndexInfo.index_id,
                            IndexInfo.desc_name,
                            IndexInfo.em_id,
                        ).filter(
                            IndexInfo.index_id.in_(index_list)
                        )
                df1 = pd.read_sql(query.statement, query.session.bind)
                dic = {
                    'real_id':'index_id',
                    'real_name':'desc_name',
                    'asset_id':'em_id',
                }
                query = db_session.query(
                            AssetInfo.asset_id,
                            AssetInfo.real_name,
                            AssetInfo.real_id,
                        ).filter(
                            AssetInfo.real_id.in_(index_list)
                        )
                df2 = pd.read_sql(query.statement, query.session.bind)
                df2 = df2.rename(columns=dic)[dic.values()]
                
            with TestDatabaseConnector().managed_session() as db_session:
                query = db_session.query(
                            PfIndexInfo.index_id,
                            PfIndexInfo.desc_name,       
                        ).filter(
                            PfIndexInfo.index_id.in_(index_list)
                        )
                df3 = pd.read_sql(query.statement, query.session.bind)
            df = pd.concat([df1,df2,df3])
            return df
                
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from {ResearchApi.get_index_info}')


    def get_index_component(self,index_id,start_date,end_date):
        '''指数成分股查询'''
        with RawDatabaseConnector().managed_session() as db_session:
            try: 
                    query = db_session.query(
                        EmIndexComponent
                    ).filter(
                        EmIndexComponent.index_id.in_([index_id]),
                        EmIndexComponent.datetime <= end_date, 
                        EmIndexComponent.datetime >= start_date
                        )
                    index_stocks = pd.read_sql(query.statement, query.session.bind)
                    return index_stocks                  
            except Exception as e:
                    print(f'Failed to get data <err_msg> {e} from {ResearchApi.get_index_component}')

    def get_index_fac(self, start_date, end_date, index_ids:list=[]):
        '''指数衍生数据查询接口'''
        with DerivedDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                            IndexValuationLongTerm.index_id,
                            IndexValuationLongTerm.datetime,
                            IndexValuationLongTerm.pb_mrq,
                            IndexValuationLongTerm.pe_ttm,
                            IndexValuationLongTerm.roe,
                            IndexValuationLongTerm.ps_ttm,
                            IndexValuationLongTerm.dy,
                            IndexValuationLongTerm.pcf_ttm,
                            IndexValuationLongTerm.est_peg,
                            IndexValuationLongTerm.eps_ttm,
                    )
 
                if index_ids:
                    query = query.filter(
                        IndexValuationLongTerm.index_id.in_(index_ids),
                    )
                if start_date:
                    query = query.filter(
                        IndexValuationLongTerm.datetime >= start_date,
                    )
                if end_date:
                    query = query.filter(
                        IndexValuationLongTerm.datetime <= end_date,
                    )
                df = pd.read_sql(query.statement, query.session.bind)
                return df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {ResearchApi.get_index_valuation_long_term}')                    
                    
    def get_fund_nav(self , start_date , end_date , fund_id_list):
        '''公募净值查询接口'''
        try:
            with BasicDatabaseConnector().managed_session() as db_session:
                rename_dic = {
                    'adjusted_net_value': 'unitnav',
                    'acc_net_value':'cumnav',
                    'adjusted_net_value':'adjnav',
                    'daily_profit':'mmf_unityield',
                    'PURCHSTATUS':'purchase_status',
                    'REDEMSTATUS':'redeem_status',
                }
                query = db_session.query(
                    FundNav.fund_id,
                    FundNav.datetime,
                    FundNav.adjusted_net_value,
                    FundNav.unit_net_value,
                    FundNav.acc_net_value,
                    FundNav.daily_profit,
                ).filter(
                    FundNav.fund_id.in_(fund_id_list),
                    FundNav.datetime >= start_date,
                    FundNav.datetime <= end_date,
                )
                df = pd.read_sql(query.statement, query.session.bind)
                res = []
                for fund_id in fund_id_list:
                    df_i = df[df.fund_id == fund_id].copy()
                    df_i.loc[:,'ret'] = df_i.adjusted_net_value.pct_change(1)
                    res.append(df_i)
                df = pd.concat(res).rename(columns=rename_dic).reset_index(drop=True)
                fund_ids = df.fund_id.unique().tolist()
                fund_ids = [fund_id.split('!')[0] for fund_id in fund_ids]

            with RawDatabaseConnector().managed_session() as db_session:     

                query = db_session.query(
                        EmFundStatus.CODES,
                        EmFundStatus.DATES,
                        EmFundStatus.PURCHSTATUS,
                        EmFundStatus.REDEMSTATUS
                    )
                cons = (EmFundStatus.CODES.like(f'%{fund_id}%') for fund_id in fund_ids)
                query = query.filter(or_(cons))
                fund_status = pd.read_sql(query.statement, query.session.bind)
                df.loc[:,'_fund_id'] = df.fund_id.map(lambda x :x.split('!')[0]) 
                fund_status.loc[:,'_fund_id'] = fund_status.CODES.map(lambda x: x.split('.')[0])
                fund_ids = df._fund_id.unique().tolist()
                res = []
                for fund_id in fund_ids:
                    df_x = df[df._fund_id == fund_id].copy()
                    df_y = fund_status[fund_status._fund_id == fund_id].copy().rename(columns={'DATES':'datetime'})[['datetime','PURCHSTATUS','REDEMSTATUS']]
                    df_i = pd.merge(df_x, df_y, on='datetime',how='outer').set_index('datetime').sort_index()
                    df_i[['PURCHSTATUS','REDEMSTATUS']] = df_i[['PURCHSTATUS','REDEMSTATUS']].ffill()
                    df_i = df_i.dropna(subset=['fund_id']).drop(labels=['_fund_id'], axis=1)
                    res.append(df_i)
                df = pd.concat(res).rename(columns=rename_dic).reset_index()
                return df
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from ResearchApi.get_fund_nav')
    
    def get_fund_asset_alloc(self , start_date , end_date , fund_id_list):
        '''公募基金持仓资产查询接口'''
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundHoldAsset.fund_id,
                    FundHoldAsset.datetime,
                    FundHoldAsset.stock_nav_ratio,
                    FundHoldAsset.bond_nav_ratio,
                    FundHoldAsset.fund_nav_ratio,
                    FundHoldAsset.cash_nav_ratio,
                    FundHoldAsset.other_nav_ratio,
                    FundHoldAsset.first_repo_to_nav,
                    FundHoldAsset.avg_ptm,
                    FundHoldAsset._update_time,
                ).filter(
                    FundHoldAsset.fund_id.in_(fund_id_list),
                    FundHoldAsset.datetime >= start_date,
                    FundHoldAsset.datetime <= end_date,
                )
                df = pd.read_sql(query.statement, query.session.bind)
                dts = pd.to_datetime(df._update_time)
                dts = [i.date() for i in dts]
                df['publish_date'] = dts
                df = df.drop(labels=['_update_time'], axis=1)
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from ResearchApi.get_fund_hold_asset')
                
    def get_fund_hold_stock(self , start_date , end_date , fund_id_list):
        '''公募基金持仓股票查询接口'''
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FundHoldStock
                ).filter(
                    FundHoldStock.fund_id.in_(fund_id_list),
                    FundHoldStock.datetime >= start_date,
                    FundHoldStock.datetime <= end_date,
                )
                df = pd.read_sql(query.statement, query.session.bind).drop(columns = ["_update_time"])
                df = df.dropna(subset=['rank1_stock']).reset_index(drop=True)
                common_cols = ['fund_id','datetime']
                res = []
                for idx in df.index:
                    for i in range(1,11):
                        rank_cols = [f'rank{i}_stock',f'rank{i}_stock_code',f'rank{i}_stockval',f'rank{i}_stockweight']
                        rank_dic = {_i:_i.replace(str(i),'') for _i in rank_cols}
                        _df = df.loc[[idx]][common_cols+rank_cols].rename(columns=rank_dic).copy()
                        _df['rank'] = str(i)
                        res.append(_df)
                df = pd.concat(res)
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from ResearchApi.get_fund_hold_stock')

    def get_fund_info(self, fund_id = None ,desc_name = None ,wind_class_1 = None ,wind_class_2 = None ,
                  company_id = None, start_date = None ,size = None ,manager_id = None):
        """公募产品基本信息查询接口"""
        
        def obscure_return(Connector,list_,col_name,que):
            """
            支持模糊查询的方法
            Connector : databse managed session
            list_ : 一个含有模糊或者准确信息字符组成的list
            col_name: 筛选条件对应在databse的columns name
            return 一个含有准确信息字符串组成的list
            """
            with Connector as db_session:
                            query = db_session.query(
                                    que
                                )       
                            df = pd.read_sql(query.statement, query.session.bind)
            bool_df = pd.DataFrame()
            for str_ in list_:
                bool_df[str_] = (df[col_name].str.contains(str_))
            result = list(set(df[bool_df.any(axis = 1)][col_name].values.tolist()))
            return result
    
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query_size = db_session.query(
                                Fund_size_and_hold_rate.size,
                                Fund_size_and_hold_rate.fund_id                           
                            )
                df_size = pd.read_sql(query_size.statement, query_size.session.bind)
                query = db_session.query(
                            FundInfo.fund_id,
                            FundInfo.desc_name,
                            FundInfo.wind_class_1,
                            FundInfo.wind_class_2,
                            FundInfo.company_id,
                            FundInfo.start_date,
                            FundInfo.manager_id, 
                            FundInfo.end_date,
                        )
                if(fund_id != None):
                    query = query.filter(
                        FundInfo.fund_id.in_(fund_id),
                    )
                if(desc_name != None):
                    desc_name = obscure_return(BasicDatabaseConnector().managed_session(),desc_name,
                                                       "desc_name",FundInfo.desc_name)#模糊查询
                    query = query.filter(
                        FundInfo.desc_name.in_(desc_name),
                    )
                if(wind_class_1 != None):
                    wind_class_1 = obscure_return(BasicDatabaseConnector().managed_session(),wind_class_1,
                                                       "wind_class_1",FundInfo.wind_class_1)
                    query = query.filter(
                        FundInfo.wind_class_1.in_(wind_class_1),
                    )
                if(wind_class_2 != None):
                    wind_class_2 = obscure_return(BasicDatabaseConnector().managed_session(),wind_class_2,
                                                       "wind_class_2",FundInfo.wind_class_2)
                    query = query.filter(
                        FundInfo.wind_class_2.in_(wind_class_2),
                    )
                if(company_id != None):
                    company_id = obscure_return(BasicDatabaseConnector().managed_session(),company_id,
                                                      "company_id",FundInfo.company_id)#模糊查询
                    query = query.filter(
                        FundInfo.company_id.in_(company_id),
                    )
                if(start_date != None):
                    query = query.filter(
                        FundInfo.start_date >= start_date
                    )
                if(manager_id != None):
                    manager_id = obscure_return(BasicDatabaseConnector().managed_session(),manager_id,
                                                      "manager_id",FundInfo.manager_id)#模糊查询
                    query = query.filter(
                        FundInfo.manager_id.in_(manager_id),
                    )
                    
                df = pd.read_sql(query.statement, query.session.bind)
                df_size = df_size.dropna(subset=['size'],axis = 0)
                df_size = df_size.drop_duplicates(subset='fund_id', keep="last")#这里选取最新的基金规模             
                result_df = pd.merge(df,df_size,how = 'left',on = 'fund_id') 
                if (size != None):
                    result_df = result_df[(result_df['size'] > size[0]) & (result_df['size'] < size[1])]
                return result_df
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e) + " get_fund_info()")
                return None

    def get_stock_quote(self , start_date , end_date , stock_id_list):
        '''股票行情查询接口'''
        with RawDatabaseConnector().managed_session() as db_session:
            try:
                #EmStockPrice
                post = 'post_'
                rename_dic = {
                    'open':f'{post}open',
                    'close':f'{post}close',
                    'high':f'{post}high',
                    'low':f'{post}low',
                }
                query = db_session.query(
                    EmStockPostPrice.CODES,
                    EmStockPostPrice.DATES,
                    EmStockPostPrice.OPEN,
                    EmStockPostPrice.CLOSE,
                    EmStockPostPrice.HIGH,
                    EmStockPostPrice.LOW,
                    EmStockPostPrice.TRADESTATUS
                ).filter(
                    EmStockPostPrice.CODES.in_(stock_id_list),
                    EmStockPostPrice.DATES >= start_date-datetime.timedelta(days=30),
                    EmStockPostPrice.DATES <= end_date,
                )
                df = pd.read_sql(query.statement, query.session.bind).set_index('datetime')
                df = df.rename(columns=rename_dic)
                query = db_session.query(
                    EmStockPrice.CODES,
                    EmStockPrice.DATES,
                    EmStockPrice.OPEN,
                    EmStockPrice.CLOSE,
                    EmStockPrice.HIGH,
                    EmStockPrice.LOW,
                    EmStockPrice.VOLUME,
                    EmStockPrice.AMOUNT,
                ).filter(
                    EmStockPrice.CODES.in_(stock_id_list),
                    EmStockPrice.DATES >= start_date,
                    EmStockPrice.DATES <= end_date,
                )
                _df = pd.read_sql(query.statement, query.session.bind).set_index('datetime')

                res = []
                for code in df.stock_id.unique():
                    df_i = df[df.stock_id == code].copy()
                    df_i['ret'] = df_i.post_close.pct_change(1)
                    df_i = df_i.loc[start_date:]
                    _df_i = _df[_df.stock_id == code][['open','close','high','low','volume','amount']]
                    df_i = df_i.join(_df_i).reset_index()
                    res.append(df_i)
                df = pd.concat(res)
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from ResearchApi.get_stock_quote')
                
    def get_stock_post_price(self,start_date=None,end_date=None,stock_list=None):
         with RawDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    EmStockPostPrice.CODES,
                    EmStockPostPrice.DATES,
                    EmStockPostPrice.OPEN,
                    EmStockPostPrice.CLOSE,
                    EmStockPostPrice.HIGH,
                    EmStockPostPrice.LOW,
                    EmStockPostPrice.TRADESTATUS
                )
                if start_date:
                    query = query.filter(EmStockPostPrice.DATES >= start_date)
                if end_date:
                    query = query.filter(EmStockPostPrice.DATES <= end_date)
                if stock_list:
                    query = query.filter(EmStockPostPrice.CODES.in_(stock_list))
                df = pd.read_sql(query.statement, query.session.bind).set_index('datetime')
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from ResearchApi.get_stock_post_price')

    def get_macro_value(self, code_list, start_date, end_date):
        '''宏观数据获取'''
        with RawDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        EmMacroEconomicMonthly.codes,
                        EmMacroEconomicMonthly.datetime,
                        EmMacroEconomicMonthly.value,
                    ).filter(
                        EmMacroEconomicMonthly.codes.in_(code_list),
                        EmMacroEconomicMonthly.datetime >= start_date,
                        EmMacroEconomicMonthly.datetime <= end_date,
                    )
                df1 = pd.read_sql(query.statement, query.session.bind).drop(columns = "_update_time", errors='ignore')
                query = db_session.query(
                        EmMacroEconomicDaily.codes,
                        EmMacroEconomicDaily.datetime,
                        EmMacroEconomicDaily.value,
                    ).filter(
                        EmMacroEconomicDaily.codes.in_(code_list),
                        EmMacroEconomicDaily.datetime >= start_date,
                        EmMacroEconomicDaily.datetime <= end_date,
                    )
                df2 = pd.read_sql(query.statement, query.session.bind).drop(columns = "_update_time", errors='ignore')
                df = pd.concat([df1, df2])
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from ResearchApi.get_macro_value')

    def get_macro_info(self, code_list=None, desc_list=None):
        '''宏观信息'''
        with RawDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        EmMacroEconomicInfo.codes,
                        EmMacroEconomicInfo.desc_name,
                    )
                if code_list:
                    query = query.filter(
                            EmMacroEconomicInfo.codes.in_(code_list),   
                        )
                if desc_list:    
                    for i in desc_list:
                        query = query.filter(
                            EmMacroEconomicInfo.desc_name.like(f'%{i}%')
                        )
                df = pd.read_sql(query.statement, query.session.bind)
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from ResearchApi.get_macro_info')

    def get_future_info(self, future_ids=None, trans_type_list=None):
        '''期货信息'''
        with RawDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        EmFutureInfoDetail.EMCODE,
                        EmFutureInfoDetail.TRADINGCODE,
                        EmFutureInfoDetail.FTTRANSTYPE,
                        EmFutureInfoDetail.CONTRACTMUL,
                        EmFutureInfoDetail.NAME,
                        )
                if future_ids:
                    query = query.filter(
                            EmFutureInfoDetail.EMCODE.in_(future_ids),   
                        )
                if trans_type_list:    
                    for i in trans_type_list:
                        query = query.filter(
                            EmFutureInfoDetail.FTTRANSTYPE.like(f'%{i}%')
                        )
                df = pd.read_sql(query.statement, query.session.bind)
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from ResearchApi.get_future_info')



    def get_future_quote(self, future_ids=None, start_date=None, end_date=None):
        '''期货行情'''
        with RawDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        EmFuturePriceDetail.CODES,
                        EmFuturePriceDetail.DATES,
                        EmFuturePriceDetail.OPEN,
                        EmFuturePriceDetail.HIGH,
                        EmFuturePriceDetail.LOW,
                        EmFuturePriceDetail.CLOSE,
                        EmFuturePriceDetail.PRECLOSE,
                        EmFuturePriceDetail.AVERAGE,
                        EmFuturePriceDetail.VOLUME,
                        EmFuturePriceDetail.AMOUNT,
                        EmFuturePriceDetail.HQOI,
                        EmFuturePriceDetail.CLEAR,
                        EmFuturePriceDetail.PRECLEAR,
                    )
                if future_ids:
                    query = query.filter(
                        EmFuturePriceDetail.CODES.in_(future_ids),
                    )
                if start_date:
                    query = query.filter(
                        EmFuturePriceDetail.DATES >= start_date,
                    )
                if end_date:
                    query = query.filter(
                        EmFuturePriceDetail.DATES <= end_date,
                    )
                df = pd.read_sql(query.statement, query.session.bind).drop(columns = "_update_time", errors='ignore')
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from ResearchApi.get_macro_value')

    def get_industry_info(self, industry_class):
        '''行业分类详情查询接口'''
        try:
            if industry_class == "sw":
                with RawDatabaseConnector().managed_session() as db_session:
                    query = db_session.query(
                        EmIndustryInfo
                    ).filter(
                    )
                    df = pd.read_sql(query.statement, query.session.bind).drop(columns = "_update_time", errors='ignore')
                with BasicDatabaseConnector().managed_session() as db_session:
                    query = db_session.query(
                        IndexInfo.em_id,
                        IndexInfo.index_id
                    ).filter(
                        IndexInfo.em_id.in_(df.em_id)
                    )
                    df_1 = pd.read_sql(query.statement, query.session.bind).drop(columns = "_update_time", errors='ignore')
                    df = pd.merge(df, df_1, on='em_id')
                    return df
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from ResearchApi.get_industry_info')
            
    def get_stock_info_concept(self, stock_id_list=None):
        '''股票申万\中信行业分类查询接口'''
        with RawDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    EmStockConceptInfo
                )
                if stock_id_list:
                    query = query.filter(
                        EmStockConceptInfo.codes.in_(stock_id_list),
                    )
                df = pd.read_sql(query.statement, query.session.bind).drop(columns = "_update_time").set_index("codes")
                
                query = db_session.query(
                    EmStockInfo
                )
                if stock_id_list:
                    query = query.filter(
                        EmStockInfo.CODES.in_(stock_id_list),
                    )
                df = df.join(pd.read_sql(query.statement, query.session.bind).drop(columns = "_update_time").set_index("stock_id")[["bl_sws_ind_code"]])
                df['bl_sws_ind_code'] = df['bl_sws_ind_code'].map(lambda x: x.split('-') if isinstance(x, str) else None)
                df = df.explode('bl_sws_ind_code')
                query = db_session.query(
                    EmIndustryInfo.em_id,
                    EmIndustryInfo.ind_name,
                ).filter(
                    EmIndustryInfo.em_id.in_(df.bl_sws_ind_code.unique().tolist())
                )
                _df = pd.read_sql(query.statement, query.session.bind).rename(columns={'em_id':'bl_sws_ind_code'})
                df = pd.merge(df.reset_index(),_df,on='bl_sws_ind_code')
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from ResearchApi.get_stock_info_concept')
                
    def get_stock_fin(self , start_date , end_date , stock_id_list):
        '''股票财务数据查询接口'''
        with RawDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    EmStockFinFac.CODES,
                    EmStockFinFac.DATES,
                    EmStockFinFac.ROEAVG,
                    EmStockFinFac.EPSDILUTED,
                    EmStockFinFac.BPS,
                    EmStockFinFac._update_time,
                ).filter(
                    EmStockFinFac.DATES >= start_date,
                    EmStockFinFac.DATES <= end_date,
                    EmStockFinFac.CODES.in_(stock_id_list),
                )
                df = pd.read_sql(query.statement, query.session.bind)
                dts = pd.to_datetime(df._update_time)
                dts = [i.date() for i in dts]
                df['publish_date'] = dts
                df = df.drop(labels=['_update_time'], axis=1)
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from ResearchApi.get_stock_price')
                
                
    def get_stock_fin_fac(self , start_date , end_date , stock_id_list):
        '''股票衍生数据查询接口'''
        with RawDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    EmStockDailyInfo.CODES,
                    EmStockDailyInfo.DATES,
                    EmStockDailyInfo.PETTMDEDUCTED,
                    EmStockDailyInfo.PBLYRN,
                    EmStockDailyInfo.PSTTM,
                    EmStockDailyInfo.ESTPEG,
                    EmStockDailyInfo.EVWITHOUTCASH,
                ).filter(
                    EmStockDailyInfo.CODES.in_(stock_id_list),
                    EmStockDailyInfo.DATES >= start_date,
                    EmStockDailyInfo.DATES <= end_date,
                )
                df = pd.read_sql(query.statement, query.session.bind)#.drop(columns = "_update_time")
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from ResearchApi.get_stock_price')
                
                
    def get_capital_north_south(self , capital_class , start_date , end_date):
        '''北上\南下资金数据查询接口'''
        if capital_class == "south":
            with RawDatabaseConnector().managed_session() as db_session:
                try:
                    query = db_session.query(
                        EMSouthCapital
                    ).filter(
                        EMSouthCapital.datetime >= start_date,
                        EMSouthCapital.datetime <= end_date,
                    )
                    df = pd.read_sql(query.statement, query.session.bind).drop(columns = "_update_time")
                    return df
                except Exception as e:
                    print(f'Failed to get data <err_msg> {e} from ResearchApi.get_stock_price')
        else:
            with RawDatabaseConnector().managed_session() as db_session:
                try:
                    query = db_session.query(
                        EMNorthCapital
                    ).filter(
                        EMNorthCapital.datetime >= start_date,
                        EMNorthCapital.datetime <= end_date,
                    )
                    df = pd.read_sql(query.statement, query.session.bind).drop(columns = "_update_time")
                    return df
                except Exception as e:
                    print(f'Failed to get data <err_msg> {e} from ResearchApi.get_stock_price')
                
    
    def get_pf_info(self,
                    record_cd=None, 
                    name_list=None, 
                    stg_name=None, 
                    stg_child=None, 
                    company_name=None,
                    start_date=None,
                    mng_name=None):
        '''私募产品基本信息查询接口'''
        with TLDatabaseConnector().managed_session() as db_session:
            # 主策略对应
            main_type_name = '私募基金投资策略' 
            _query = db_session.query(
                SysCode.VALUE_NAME_CN,
                SysCode.VALUE_NUM_CD,
            ).filter(
                SysCode.CODE_TYPE_NAME == main_type_name
            )
            df = pd.read_sql(_query.statement, _query.session.bind)
            main_stg_dic = df.set_index('VALUE_NAME_CN')['VALUE_NUM_CD'].to_dict()
            main_stg_dic_re = df.set_index('VALUE_NUM_CD')['VALUE_NAME_CN'].to_dict()
            # 子策略对应
            child_type_name = '私募基金投资子策略'
            query = db_session.query(
                SysCode.VALUE_NAME_CN,
                SysCode.VALUE_NUM_CD,
            ).filter(
                SysCode.CODE_TYPE_NAME == child_type_name
            )
            df = pd.read_sql(query.statement, query.session.bind)
            child_stg_dic = df.set_index('VALUE_NAME_CN')['VALUE_NUM_CD'].to_dict()
            child_stg_dic_re = df.set_index('VALUE_NUM_CD')['VALUE_NAME_CN'].to_dict()

            # 返回字段
            query = db_session.query(
                Pfund.RECORD_CD,
                Pfund.INVEST_STRATEGY,
                Pfund.INVEST_STRATEGY_CHILD,
                Pfund.INVEST_CONSULTANT,
                Pfund.ESTABLISH_DATE,
                Pfund.END_DATE,
                Pfund.MANAGER,
                Pfund.STATUS,
                Pfund.SECURITY_ID,
            )
            # 按照备案号
            if record_cd:
                query = query.filter(
                    Pfund.RECORD_CD.in_(record_cd),
                )
            # 按照关键字搜索
            if name_list:
                _query = db_session.query(
                    MdSecurity.SECURITY_ID,
                    MdSecurity.SEC_SHORT_NAME,
                ).filter(MdSecurity.ASSET_CLASS=='FP')
                for i in name_list:
                    _query = _query.filter(
                        MdSecurity.SEC_SHORT_NAME.like(f'%{i}%')
                    )
                asset_info_df = pd.read_sql(_query.statement, _query.session.bind)
                query = query.filter(
                    Pfund.SECURITY_ID.in_(asset_info_df.SECURITY_ID),
                )
            # 按照主策略
            if stg_name:

                stg_list = [main_stg_dic.get(i) for i in stg_name]
                query = query.filter(
                    Pfund.INVEST_STRATEGY.in_(stg_list),
                )
            # 按照子策略
            if stg_child:

                stg_list = [child_stg_dic.get(i) for i in stg_child]
                query = query.filter(
                    Pfund.INVEST_STRATEGY_CHILD.in_(stg_list),
                )
            # 按照公司名搜索
            if company_name:
                _query = db_session.query(
                    MdInstitution.PARTY_ID,
                    MdInstitution.PARTY_SHORT_NAME,
                )
                for i in company_name:
                    _query = _query.filter(
                        MdInstitution.PARTY_SHORT_NAME.like(f'%{i}%')
                    )
                asset_info_df = pd.read_sql(_query.statement, _query.session.bind)
                query = query.filter(
                    Pfund.INVEST_CONSULTANT.in_(asset_info_df.PARTY_ID.dropna().unique()),
                )    
            # 按照成立日期
            if start_date:
                query = query.filter(
                    Pfund.ESTABLISH_DATE > start_date,
                ) 
            # 按照基金经理
            if mng_name:
                cons = (Pfund.MANAGER.like(f'%{i}%') for i in mng_name)
                query = query.filter(or_(cons))
            df = pd.read_sql(query.statement, query.session.bind).dropna(subset=['RECORD_CD'])
            df.loc[:,'INVEST_STRATEGY'] = df.INVEST_STRATEGY.map(main_stg_dic_re)
            df.loc[:,'INVEST_STRATEGY_CHILD'] = df.INVEST_STRATEGY_CHILD.map(child_stg_dic_re)
            # 替换机构名
            _query = db_session.query(
                MdInstitution.PARTY_ID,
                MdInstitution.PARTY_SHORT_NAME,
            ).filter(MdInstitution.PARTY_ID.in_(df.INVEST_CONSULTANT))
            _df = pd.read_sql(_query.statement, _query.session.bind).dropna(subset=['PARTY_ID'])
            _df['PARTY_ID'] = _df['PARTY_ID'].astype(str)
            company_dic = _df.set_index('PARTY_ID')['PARTY_SHORT_NAME'].to_dict()
            df.loc[:,'INVEST_CONSULTANT'] = df.INVEST_CONSULTANT.map(lambda x: company_dic.get(x))

            # 替换基金全名
            _query = db_session.query(
                MdSecurity.SECURITY_ID,
                MdSecurity.SEC_SHORT_NAME,
            ).filter(MdSecurity.SECURITY_ID.in_(df.SECURITY_ID))
            _df = pd.read_sql(_query.statement, _query.session.bind)
            name_dic = _df.set_index('SECURITY_ID')['SEC_SHORT_NAME'].to_dict()
            df.loc[:,'SEC_FULL_NAME'] = df.SECURITY_ID.map(name_dic)
            df = df.drop(columns=['SECURITY_ID'])

            # 替换运作状态
            _query = db_session.query(
                SysCode.VALUE_NUM_CD,
                SysCode.VALUE_NAME_CN,
            ).filter(SysCode.CODE_TYPE_ID == '40034' )
            _df = pd.read_sql(_query.statement, _query.session.bind)
            name_dic = _df.set_index('VALUE_NUM_CD')['VALUE_NAME_CN'].to_dict()
            df['STATUS'] = df.STATUS.map(name_dic)
            return df

    def get_pf_info2(self, record_cd):
        '''私募产品要素查询接口''' 
        with TLDatabaseConnector().managed_session() as db_session:    
            # 返回字段
            query = db_session.query(
                Pfund.RECORD_CD,
                Pfund.INVEST_CONSULTANT,
                Pfund.CUSTODIAN,
                Pfund.SUBSCRIPTION_START_POINT,
                Pfund.MIN_ADD,
                Pfund.STOP_LOSS_LINE,
                Pfund.WARN_LINE,
                Pfund.ISSUE_FEE,
                Pfund.REDEEM_FEE,
                Pfund.MANAGEMENT_FEE,
                Pfund.PERFORMANECE_RETURN,
                Pfund.APPLY_FEE,
                Pfund.CUSTODY_FEE,
                Pfund.SECURITY_ID,
            ).filter(
                Pfund.RECORD_CD.in_(record_cd),
            )
            df = pd.read_sql(query.statement, query.session.bind)
            _query = db_session.query(
                MdSecurity.SECURITY_ID,
                MdSecurity.SEC_SHORT_NAME,
            ).filter(MdSecurity.SECURITY_ID.in_(df.SECURITY_ID))
            _df = pd.read_sql(_query.statement, _query.session.bind)
            name_dic = _df.set_index('SECURITY_ID')['SEC_SHORT_NAME'].to_dict()
            df.loc[:,'SEC_FULL_NAME'] = df.SECURITY_ID.map(name_dic)
            df = df.drop(columns=['SECURITY_ID'])
            return df

    def search_pf_info(self, record_cd=None,start_date=None,info_items=None,with_detail=False):
        '''私募信息搜索 '''
        try:
            res = [self.get_pf_info(record_cd=record_cd,name_list=info_i,start_date=start_date) for info_i in info_items ]
            df_name = pd.concat(res).drop_duplicates()

            res = [self.get_pf_info(record_cd=record_cd,stg_name=info_i,start_date=start_date) for info_i in info_items ]
            df_stg = pd.concat(res).drop_duplicates()

            res = [self.get_pf_info(record_cd=record_cd,stg_child=info_i,start_date=start_date) for info_i in info_items ]
            df_stg_chd = pd.concat(res).drop_duplicates()

            time.sleep(1) # 搜索量大， 抱错
            res = [self.get_pf_info(record_cd=record_cd,company_name=info_i,start_date=start_date) for info_i in info_items ]
            df_stg_cpy = pd.concat(res).drop_duplicates()

            res = [self.get_pf_info(record_cd=record_cd,mng_name=info_i,start_date=start_date) for info_i in info_items ]
            df_stg_mng = pd.concat(res).drop_duplicates()

            df = pd.concat([df_name,df_stg,df_stg_chd,df_stg_cpy,df_stg_mng]).drop_duplicates().reset_index(drop=True)
            if with_detail:
                _df = self.get_pf_info2(df.RECORD_CD)
                df = pd.merge(df, _df, on='RECORD_CD', how='outer')
            return df
        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from ResearchApi.search_pf_info')

    def get_pf_inst_info(self, reg_cd=None,inst_name=None,key_person=None,scale_list=None):
        '''私募投顾基本信息查询接口'''
        inst_info_df = None
        inst_scale_df = None
        # 返回字段
        with TLDatabaseConnector().managed_session() as db_session:    
            # 返回字段
            query = db_session.query(
                PfundInstInfo.REG_CD,
                PfundInstInfo.PROFILE,
                PfundInstInfo.IDEA_STRATEGY,
                PfundInstInfo.KEY_PERSON,
                PfundInstInfo.PARTY_ID,
            )
        # 按照备案号
        if reg_cd:
            query = query.filter(
                PfundInstInfo.REG_CD.in_(reg_cd),
            )
        # 按照投顾名
        if inst_name:
            _query = db_session.query(
                PfundInstScaleAmac.PARTY_ID,
                PfundInstScaleAmac.PARTY_FULL_NAME,
            )
            for i in inst_name:
                _query = _query.filter(
                    PfundInstScaleAmac.PARTY_FULL_NAME.like(f'%{i}%')
                )
            inst_info_df = pd.read_sql(_query.statement, _query.session.bind)
            query = query.filter(
                PfundInstInfo.PARTY_ID.in_(inst_info_df.PARTY_ID.dropna().unique()),
            )    
        # 核心人物
        if key_person:
            for i in key_person:
                query = query.filter(
                    PfundInstInfo.KEY_PERSON.like(f'%{i}%'),
                )
        # 按照规模
        if scale_list:
            _query = db_session.query(
                PfundInstScaleAmac.PARTY_ID,
                PfundInstScaleAmac.SCALE,
            ).filter(
                PfundInstScaleAmac.SCALE.in_(scale_list)
            )
            inst_scale_df = pd.read_sql(_query.statement, _query.session.bind)
            query = query.filter(
                PfundInstInfo.PARTY_ID.in_(inst_scale_df.PARTY_ID.dropna().unique()),
            )    
        df = pd.read_sql(query.statement, query.session.bind)
        # 添加机构名
        if inst_info_df is None:
            _query = db_session.query(
                PfundInstScaleAmac.PARTY_ID,
                PfundInstScaleAmac.PARTY_FULL_NAME,
            ).filter(
                    PfundInstScaleAmac.PARTY_ID.in_(df.PARTY_ID.dropna().unique())
            )
            inst_info_df = pd.read_sql(_query.statement, _query.session.bind)
        full_name_dic = inst_info_df.dropna(subset=['PARTY_ID']).set_index('PARTY_ID')['PARTY_FULL_NAME'].to_dict()
        df.loc[:,'PARTY_FULL_NAME'] = df.PARTY_ID.map(full_name_dic)
        
        # 添加规模
        if inst_scale_df is None:
            _query = db_session.query(
                PfundInstScaleAmac.PARTY_ID,
                PfundInstScaleAmac.SCALE,
            ).filter(
                    PfundInstScaleAmac.PARTY_ID.in_(df.PARTY_ID.dropna().unique())
            )
            inst_scale_df = pd.read_sql(_query.statement, _query.session.bind)
        scale_dic = inst_scale_df.dropna(subset=['PARTY_ID']).set_index('PARTY_ID')['SCALE'].to_dict()
        df.loc[:,'SCALE'] = df.PARTY_ID.map(scale_dic)
        # 主策略统计
        main_type_name = '私募基金投资策略' 
        _query = db_session.query(
            SysCode.VALUE_NAME_CN,
            SysCode.VALUE_NUM_CD,
        ).filter(
            SysCode.CODE_TYPE_NAME == main_type_name
        )
        sys_df = pd.read_sql(_query.statement, _query.session.bind)
        #main_stg_dic = _df.set_index('VALUE_NAME_CN')['VALUE_NUM_CD'].to_dict()
        main_stg_dic_re = sys_df.set_index('VALUE_NUM_CD')['VALUE_NAME_CN'].to_dict()
        query = db_session.query(
                Pfund.RECORD_CD,
                Pfund.INVEST_STRATEGY,    
            ).filter(
            Pfund.INVEST_CONSULTANT.in_(df.PARTY_ID.unique()),
        )    
        _df = pd.read_sql(query.statement, query.session.bind)
        _df.loc[:,'VALUE_NAME_CN'] = _df.INVEST_STRATEGY.map(main_stg_dic_re)
        if not df.empty:
            main_stg = _df.groupby('VALUE_NAME_CN').count().sort_values('RECORD_CD').index[-1]
            df.loc[:,'MAIN_FUND_TYPE'] = main_stg
        else:
            df['MAIN_FUND_TYPE'] = ''
        return df

    def get_pf_nav(self, start_date=None, end_date=None, record_cds:list=[], source='tl'):
        '''私募净值查询接口'''
        try:
            # 通联
            if source == 'tl':
                with TLDatabaseConnector().managed_session() as db_session:                
                    query1 = db_session.query(
                        Pfund.SECURITY_ID,
                        Pfund.RECORD_CD,
                        ).filter(
                        Pfund.RECORD_CD.in_(record_cds)
                        )           
                    df1 = pd.read_sql(query1.statement, query1.session.bind)
                    security_ids = df1['SECURITY_ID'].values.tolist()
                    query = db_session.query(
                            PfundNav.END_DATE,
                            PfundNav.NAV,
                            PfundNav.ADJ_NAV,
                            PfundNav.ACCUM_NAV,
                            PfundNav.SECURITY_ID
                        )
                    if record_cds:
                        query = query.filter(
                            PfundNav.SECURITY_ID.in_(security_ids),
                        )
                    if start_date:
                        query = query.filter(
                            PfundNav.END_DATE >= start_date,
                        )
                    if end_date:
                        query = query.filter(
                            PfundNav.END_DATE <= end_date,
                        )
                    df = pd.read_sql(query.statement, query.session.bind)
                    df = pd.merge(df1, df, left_on='SECURITY_ID', right_on='SECURITY_ID')
            # 自己计算
            elif source == 'py_calc':
                with TLDatabaseConnector().managed_session() as db_session:                
                    query1 = db_session.query(
                        Pfund.SECURITY_ID,
                        Pfund.RECORD_CD,
                        ).filter(
                        Pfund.RECORD_CD.in_(record_cds)
                        )           
                    df1 = pd.read_sql(query1.statement, query1.session.bind)
                    security_id_dic = df1.set_index('RECORD_CD')['SECURITY_ID'].to_dict()
                with DerivedDatabaseConnector().managed_session() as db_session:
                    name_dic = {
                        's_id':'SECURITY_ID',
                        'fof_id':'RECORD_CD',
                        'datetime':'END_DATE',
                        'nav':'NAV',
                        'adjusted_nav':'ADJ_NAV',
                        'acc_net_value':'ACCUM_NAV',
                    }
                    query = db_session.query(
                            FOFNavCalc.fof_id,
                            FOFNavCalc.datetime,
                            FOFNavCalc.nav,
                            FOFNavCalc.acc_net_value,
                            FOFNavCalc.adjusted_nav,
                        ).filter(
                            FOFNavCalc.fof_id.in_(record_cds),
                            FOFNavCalc.manager_id == "py1",
                        )
                    if start_date:
                        query = query.filter(
                            FOFNavCalc.datetime >= start_date,
                        )
                    if end_date:
                        query = query.filter(
                            FOFNavCalc.datetime <= end_date,
                        )
                    df = pd.read_sql(query.statement, query.session.bind)
                    df.loc[:,'s_id'] = df.fof_id.map(security_id_dic)
                    df = df.rename(columns=name_dic)[name_dic.values()]
            # 客服发送
            elif source == 'py':
                with TLDatabaseConnector().managed_session() as db_session:                
                    query1 = db_session.query(
                        Pfund.SECURITY_ID,
                        Pfund.RECORD_CD,
                        ).filter(
                        Pfund.RECORD_CD.in_(record_cds)
                        )           
                    df1 = pd.read_sql(query1.statement, query1.session.bind)
                    security_id_dic = df1.set_index('RECORD_CD')['SECURITY_ID'].to_dict()
                with DerivedDatabaseConnector().managed_session() as db_session:
                    name_dic = {
                        's_id':'SECURITY_ID',
                        'fof_id':'RECORD_CD',
                        'datetime':'END_DATE',
                        'nav':'NAV',
                        'adjusted_nav':'ADJ_NAV',
                        'acc_net_value':'ACCUM_NAV',
                    }
                    query = db_session.query(
                            FOFNav.fof_id,
                            FOFNav.datetime,
                            FOFNav.nav,
                            FOFNav.acc_net_value,
                            FOFNav.adjusted_nav,
                        ).filter(
                            FOFNav.fof_id.in_(record_cds),
                            FOFNav.manager_id == "py1",
                        )
                    if start_date:
                        query = query.filter(
                            FOFNav.datetime >= start_date,
                        )
                    if end_date:
                        query = query.filter(
                            FOFNav.datetime <= end_date,
                        )
                    df = pd.read_sql(query.statement, query.session.bind)
                    df.loc[:,'s_id'] = df.fof_id.map(security_id_dic)
                    df = df.rename(columns=name_dic)[name_dic.values()]
            return df

        except Exception as e:
            print(f'Failed to get data <err_msg> {e} from {ResearchApi.get_pfund_nav}')

    def get_pf_info_mixed(self, manager_id: str, fof_id: str) -> pd.DataFrame:
        '''汇总数据库中私募基金的信息表'''

        # self.get_pf_info(record_cd=fund_id_list)
        aa = self.get_pf_asset_allocation(manager_id, fof_id)
        fund_id_list = list(aa[aa.asset_type == HoldingAssetType.HEDGE].fund_id.unique()) + [fof_id]
        fof_info = BasicDataApi().get_fof_info(manager_id=manager_id, fof_id_list=fund_id_list)
        fof_info = fof_info[['fof_id', 'fof_name', 'admin', 'strategy_type']].rename(columns={'fof_id': 'fund_id'})
        fof_info['fof_id'] = fof_id
        return fof_info

    def get_pf_nav_mixed(self, manager_id: str, fof_id: str) -> pd.DataFrame:
        '获取私募基金的各类净值数据'

        def _filter_nav_with_min_max_date(x, min_max_date):
            try:
                return x[~x.datetime.between(min_max_date.at[x.fof_id.array[0], 'min_date'], min_max_date.at[x.fof_id.array[0], 'max_date'])]
            except KeyError:
                return x

        aa = self.get_pf_asset_allocation(manager_id, fof_id)
        fund_id_list = list(aa[aa.asset_type == HoldingAssetType.HEDGE].fund_id.unique()) + [fof_id]
        fof_nav_public = DerivedDataApi().get_fof_nav_public(fund_id_list)
        fof_nav_tl = ResearchApi().get_pf_nav(record_cds=fund_id_list, source='tl')
        if fof_nav_public is None or fof_nav_public.empty:
            fof_nav_public = fof_nav_tl
        else:
            fof_nav_public = fof_nav_public.drop(columns=['update_time', 'create_time', 'is_deleted'])
            if fof_nav_tl is not None:
                fof_nav_tl = fof_nav_tl[['RECORD_CD', 'END_DATE', 'NAV', 'ADJ_NAV', 'ACCUM_NAV']].rename(columns={
                    'RECORD_CD': 'fof_id',
                    'END_DATE': 'datetime',
                    'NAV': 'nav',
                    'ACCUM_NAV': 'acc_net_value',
                    'ADJ_NAV': 'adjusted_nav',
                })
                fof_nav_min_max_date = fof_nav_public.groupby(by='fof_id', sort=False).apply(lambda x: pd.Series({'min_date': x.datetime.min(), 'max_date': x.datetime.max()}))
                fof_nav_tl = fof_nav_tl.groupby(by='fof_id', sort=False, group_keys=False).apply(_filter_nav_with_min_max_date, min_max_date=fof_nav_min_max_date)
                fof_nav_public = pd.concat([fof_nav_public, fof_nav_tl], ignore_index=True).sort_values(by=['fof_id', 'datetime'])
        fof_nav_public['manager_id'] = manager_id
        fof_nav = DerivedDataApi().get_fof_nav(manager_id, fund_id_list)
        if fof_nav is None or fof_nav.empty:
            fof_nav = fof_nav_public
        else:
            fof_nav = fof_nav.drop(columns=['update_time', 'create_time', 'is_deleted'])
            if fof_nav_public is not None:
                fof_nav_min_max_date = fof_nav.groupby(by='fof_id', sort=False).apply(lambda x: pd.Series({'min_date': x.datetime.min(), 'max_date': x.datetime.max()}))
                fof_nav_public = fof_nav_public.groupby(by='fof_id', sort=False, group_keys=False).apply(_filter_nav_with_min_max_date, min_max_date=fof_nav_min_max_date)
                fof_nav = pd.concat([fof_nav, fof_nav_public], ignore_index=True).sort_values(by=['manager_id', 'fof_id', 'datetime'])
        fof_nav = fof_nav[['fof_id', 'datetime', 'nav', 'acc_net_value', 'adjusted_nav']].set_index(['fof_id', 'datetime'])
        fof_nav['v_nav'] = BasicDataApi().get_hedge_fund_email_raw(manager_id, fof_id, fund_id_list).set_index(['fund_id', 'datetime'])['v_nav']
        return fof_nav.reset_index()

    def get_pf_asset_allocation(self, manager_id: str, fof_id: str) -> pd.DataFrame:
        '''获取私募基金的交易记录'''

        df = BasicDataApi().get_fof_asset_allocation(manager_id, [fof_id])
        return df[['fof_id', 'datetime', 'fund_id', 'amount', 'share', 'nav', 'asset_type', 'event_type']]

    def get_pf_positions_detail(self, manager_id: str, fof_id: str) -> Tuple[pd.DataFrame, pd.DataFrame]:
        '''获取私募基金每日持仓市值'''

        fof_nav = DerivedDataApi().get_fof_nav(manager_id, [fof_id]).set_index('datetime').sort_index()['nav']
        trading_days = self.get_trading_date(start_date=fof_nav.index.array[0], end_date=fof_nav.index.array[-1])
        fof_nav = fof_nav.reindex(index=trading_days.datetime)
        fof_sa = DerivedDataApi().get_hedge_fund_investor_pur_redemp(manager_id, [fof_id]).set_index('datetime')
        fof_sa_sub = fof_sa.loc[fof_sa.event_type.isin([FOFTradeStatus.PURCHASE, FOFTradeStatus.SUBSCRIBE])]
        fof_sa_redemp = fof_sa.loc[fof_sa.event_type.isin([FOFTradeStatus.REDEEM,])]

        cumshares = fof_sa_sub.share_changed.fillna(0).add(fof_sa_redemp.share_changed.fillna(0), fill_value=0)
        cumshares = cumshares.groupby('datetime').sum().cumsum()

        # fof MV
        fof_mv = cumshares.reindex(fof_nav.index.union(cumshares.index)).ffill() * fof_nav
        pos_df = DerivedDataApi().get_fof_position(manager_id, [fof_id]).set_index('datetime').reindex(index=trading_days.datetime)
        sub_asset_mv = {}
        sub_asset_share = {}

        def _calc_sub_asset_mv(x):
            pos = json.loads(x.position)
            df = pd.DataFrame(pos)
            if df.empty:
                return
            df = df.set_index('fund_id')
            df = df[df.asset_type == HoldingAssetType.HEDGE]
            sub_asset_mv[x.name] = df.share * df.nav
            sub_asset_share[x.name] = df.share
            return

        pos_df[['position']].apply(_calc_sub_asset_mv, axis=1)

        # 子产品MV
        sub_asset_mv_df = pd.DataFrame.from_dict(sub_asset_mv, orient='index')
        sub_asset_mv_df = sub_asset_mv_df.reindex(fof_mv.index.union(sub_asset_mv_df.index))
        sub_asset_mv_df['others_mv'] = fof_mv - sub_asset_mv_df.sum(axis=1)

        # 子产品share
        # sub_asset_share_df = pd.DataFrame.from_dict(sub_asset_share, orient='index')
        # sub_asset_share_df = sub_asset_share_df.reindex(fof_mv.index.union(sub_asset_share_df.index))

        # 子产品损益 = MV + 赎回给的现金 + 分红的现金 - 每次投入的成本
        fof_aa = BasicDataApi().get_fof_asset_allocation(manager_id, [fof_id])
        fof_aa = fof_aa[fof_aa.asset_type == HoldingAssetType.HEDGE].drop(columns='fof_id')

        # 赎回得到的现金
        cash_redemp = fof_aa.loc[fof_aa.event_type == FOFTradeStatus.REDEEM, ['datetime', 'fund_id', 'amount']]
        cash_redemp = cash_redemp.pivot_table(index='datetime', columns='fund_id', values='amount', aggfunc=np.sum)
        cash_redemp = cash_redemp.reindex(fof_nav.index).cumsum().ffill()

        # 成本
        cost = fof_aa.loc[fof_aa.event_type.isin([FOFTradeStatus.PURCHASE, FOFTradeStatus.SUBSCRIBE]), ['datetime', 'fund_id', 'amount']]
        cost = cost.pivot_table(index='datetime', columns='fund_id', values='amount', aggfunc=np.sum).cumsum().reindex(fof_nav.index).ffill()

        # 现金分红
        cash_dividend = fof_aa.loc[fof_aa.event_type==FOFTradeStatus.DIVIDEND_CASH, ['datetime', 'fund_id', 'amount']]
        if cash_dividend.empty:
            cash_dividend = None
        else:
            cash_dividend = cost.pivot_table(index='datetime', columns='fund_id', values='amount', aggfunc=np.sum).cumsum().reindex(fof_nav.index).ffill()

        if not cash_redemp.empty:
            sub_asset_pal_fixed = sub_asset_mv_df.add(cash_redemp.fillna(0), fill_value=0)
        else:
            sub_asset_pal_fixed = sub_asset_mv_df

        if cash_dividend is not None and not cash_redemp.empty:
            sub_asset_pal_fixed = sub_asset_pal_fixed.add(cash_dividend.fillna(0), fill_value=0).sub(cost.fillna(0), fill_value=0)
        else:
            sub_asset_pal_fixed = sub_asset_pal_fixed.sub(cost.fillna(0), fill_value=0)
        return sub_asset_mv_df, sub_asset_pal_fixed

    def get_trading_date(self, asset_type='cn_stock', start_date=None, end_date=None):
        from .basic import TradingDayList
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        TradingDayList
                    )
                if start_date:
                    query = query.filter(
                        TradingDayList.datetime >= start_date,
                    )
                if end_date:
                    query = query.filter(
                        TradingDayList.datetime <= end_date,
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {TradingDayList.__tablename__}')


    def get_main_ticker(self, start_date=None,end_date=None,prod_list=None):
        try:
            with DerivedDatabaseConnector().managed_session() as db_session:
                query = db_session.query(
                    FutureMainChainRet.datetime,
                    FutureMainChainRet.future_type_id,
                    FutureMainChainRet.ticker,
                )
                if prod_list:
                    _prod_list = [_.lower() for _ in prod_list]
                    prod_list = prod_list + _prod_list
                    query = query.filter(
                        FutureMainChainRet.future_type_id.in_(prod_list),
                    )
                if start_date:
                    query = query.filter(
                        FutureMainChainRet.datetime >= start_date,
                    )
                if end_date:
                    query = query.filter(
                        FutureMainChainRet.datetime <= end_date,
                    )
                return pd.read_sql(query.statement, query.session.bind)
        
        except Exception as e:
            print(f'failed to get_future_main_chain_ret <err_msg> {e} from {FutureTypeClassification.__tablename__}')

    def get_future_and_spot_quote(self,start_date=None,end_date=None,prod_list=None):
        try:
            main_ticker_info = self.get_main_ticker(start_date=start_date,end_date=end_date,prod_list=prod_list)
            close_price = self.get_future_quote(future_ids=main_ticker_info.ticker.unique().tolist(),start_date=start_date,end_date=end_date)
            close_price = close_price.pivot_table(index='datetime',columns='em_id',values='close')
            close_price.columns.name = 'ticker'
            _close_price = close_price.unstack().rename('close').to_frame()
            future_list = main_ticker_info.future_type_id.unique().tolist()
            result = []
            for future_i in future_list:
                # 双重index对齐
                _df = main_ticker_info[main_ticker_info.future_type_id == future_i].sort_values('datetime')
                _df = _df.set_index(['ticker','datetime'])
                _df = _df.join(_close_price).reset_index() 
                spot_info = RawDataApi().get_em_future_spot_info(future_type=[future_i])
                spot_df = RawDataApi().get_em_future_spot_price(spot_ids=spot_info.spot_id.tolist(),start_date=start_date,end_date=end_date)
                spot_df = spot_df.set_index('datetime')[['price']].rename(columns={'price':'spot_price'})
                _df = _df.set_index('datetime').join(spot_df)
                result.append(_df)
            result = pd.concat(result)
            return result
        except Exception as e:
            print(f'failed to get_future_and_spot_quote <err_msg> {e} from {FutureTypeClassification.__tablename__}')

    def get_main_price_with_em_id(self, begin_date=None, end_date=None, future_id_list=[]):
        '''
        获取期货主连数据 id em_id
        '''
        try:
            with DerivedDatabaseConnector().managed_session() as db_session:
                query = db_session.query(
                    FutureMainChainRet)
                if future_id_list:
                    query = query.filter(
                        FutureMainChainRet.future_type_id.in_(future_id_list),
                    )
                if begin_date:
                    query = query.filter(
                        FutureMainChainRet.datetime >= begin_date,
                    )
                if end_date:
                    query = query.filter(
                        FutureMainChainRet.datetime <= end_date,
                    )

                return pd.read_sql(query.statement, query.session.bind).drop(columns=['_update_time'])
           
        except Exception as e:
            print(f'failed to get_main_price_with_em_id <err_msg> {e}')


    def get_main_price(self,tps=None,start_date=None,end_date=None):
        '''
        获取期货主连数据及所属主力合约行情
        tps 品种类别
        id trading_id
        '''
        try:
            with DerivedDatabaseConnector().managed_session() as db_session:
                query = db_session.query(
                    FutureMainChainRet.datetime,
                    FutureMainChainRet.future_type_id,
                    FutureMainChainRet.ticker,
                    FutureMainChainRet.ret,
                )
                if tps:
                    query = query.filter(
                        FutureMainChainRet.future_type_id.in_(tps),
                    )
                if start_date:
                    query = query.filter(
                        FutureMainChainRet.datetime >= start_date,
                    )
                if end_date:
                    query = query.filter(
                        FutureMainChainRet.datetime <= end_date,
                    )
                df = pd.read_sql(query.statement, query.session.bind)
                future_id_list = df.ticker.unique().tolist()

            with RawDatabaseConnector().managed_session() as db_session:
                query = db_session.query(
                    EmFutureInfoDetail.EMCODE,
                    EmFutureInfoDetail.TRADINGCODE,
                )
                if tps:
                    query = query.filter(
                        EmFutureInfoDetail.EMCODE.in_(future_id_list)
                )
                future_info = pd.read_sql(query.statement, query.session.bind)
                future_id_list = df.ticker.unique().tolist()
                dic = future_info.set_index('em_id')['trading_code'].to_dict()
                df['ticker'] = df.ticker.map(dic)
            df = df.set_index(['datetime','ticker'])
            future_price = self.get_future_price(tps=tps,start_date=start_date,end_date=end_date)
            future_price = future_price.set_index(['datetime','ticker'])
            df = df.join(future_price).reset_index()
            tds = pd.to_datetime(df.datetime)
            tds = [i.date() for i in tds]
            df.datetime = tds
            df = df.drop(columns=['future_type_id'])
            return df

        except Exception as e:
            print(f'failed to get_main_price <err_msg> {e}')

    def get_future_price(self,tps=None,start_date=None,end_date=None):
        '''
        获取期货行情数据
        '''
        try:
            # 获取期货品种名称
            with DerivedDatabaseConnector().managed_session() as db_session:
                query = db_session.query(
                    FutureTypeClassification.future_type_id,
                    FutureTypeClassification.desc_name,
                )
                if tps:
                    query = query.filter(
                        FutureTypeClassification.future_type_id.in_(tps),
                    )
                future_type_df = pd.read_sql(query.statement, query.session.bind)
                desc_name_list = future_type_df.desc_name.unique().tolist()

            # 获取期货合约代码
            with RawDatabaseConnector().managed_session() as db_session:
                query = db_session.query(
                    EmFutureInfoDetail.EMCODE,
                    EmFutureInfoDetail.FTTRANSTYPE,
                    EmFutureInfoDetail.TRADINGCODE,
                )
                if tps:
                    query = query.filter(
                        EmFutureInfoDetail.FTTRANSTYPE.in_(desc_name_list)
                )
                dic = future_type_df.set_index('desc_name')['future_type_id'].to_dict()
                future_info = pd.read_sql(query.statement, query.session.bind)
                future_info['future_type_id'] = future_info.trans_type.map(dic)
                
                future_ids = future_info.em_id.tolist()
                query = db_session.query(
                        EmFuturePriceDetail.CODES,
                        EmFuturePriceDetail.DATES,
                        EmFuturePriceDetail.OPEN,
                        EmFuturePriceDetail.HIGH,
                        EmFuturePriceDetail.LOW,
                        EmFuturePriceDetail.CLOSE,
                        EmFuturePriceDetail.PRECLOSE,
                        EmFuturePriceDetail.AVERAGE,
                        EmFuturePriceDetail.VOLUME,
                        EmFuturePriceDetail.AMOUNT,
                        EmFuturePriceDetail.HQOI,
                        EmFuturePriceDetail.CLEAR,
                        EmFuturePriceDetail.PRECLEAR,
                    ).filter(EmFuturePriceDetail.CODES.in_(future_ids))
                if start_date:
                    query = query.filter(EmFuturePriceDetail.DATES >= start_date)
                if end_date:
                    query = query.filter(EmFuturePriceDetail.DATES <= end_date)
                df = pd.read_sql(query.statement, query.session.bind).drop(columns = "_update_time", errors='ignore')
                df['ticker'] = df.em_id.map(future_info.set_index('em_id')['trading_code'].to_dict())
                df['product'] = df.em_id.map(future_info.set_index('em_id')['future_type_id'].to_dict())

                l = df.columns.tolist()
                target_cols = ['ticker','product']
                cols = target_cols + [i for i in l if i not in target_cols]
                df = df[cols]
            return df.drop(columns=['em_id'])

        except Exception as e:
            print(f'failed to get_future_price <err_msg> {e}')

    def get_holidays(self):
        '''
            func: 读取节假日序列
            e.g. hol = get_holidays()
            out: dict, hol['sse']是上交所节假日序列
        '''
        with BasicDatabaseConnector().managed_session() as db_session:
            query = db_session.query(
                    ExchangeHoliday
                )
            data = pd.read_sql(query.statement, query.session.bind)
                    
        hol = {}
        for exch in data['exch_id'].unique():
            hol[exch] = list(data.loc[data['exch_id']==exch, 'datetime'])
        
        return hol

    def is_trading_day(self, T, exch='sse'):
        '''
            func: 判断是否交易日
            exch: 'sse', 'szse', 'nyse', 'lse', 'tse', 'hkex', 'fse'
            e.g. tmp = is_trading_day(dt.date(2020,3,11), 'sse')
            out: True or False

        '''
        return T in self.get_tdates(start_date=T, end_date=T)

    def prev_tdate(self,T, exch='sse'):
        '''
            func: 计算前一个交易日
            exch: 'sse', 'szse', 'nyse', 'lse', 'tse', 'hkex', 'fse'
            e.g. T1 = prev_tdate(dt.date(2020,3,11), 'sse')
            out: dt.date

        '''
        # 检查输入参数
        assert(exch in ['sse', 'szse', 'nyse', 'lse', 'tse', 'hkex', 'fse'])
        T1 = T + datetime.timedelta(days=-1)
        hol = self.get_holidays()
        while ((T1 in hol[exch]) or (T1.weekday() in [5, 6])):
            T1 = T1 + datetime.timedelta(days=-1)
        return T1

    def next_tdate(self, T, exch='sse'):
        '''
            func: 计算后一个交易日
            exch: 'sse', 'szse', 'nyse', 'lse', 'tse', 'hkex', 'fse'
            e.g. T1 = next_tdate(dt.date(2020,3,11), 'sse')
            out: dt.date

        '''
        # 检查输入参数
        assert(exch in ['sse', 'szse', 'nyse', 'lse', 'tse', 'hkex', 'fse'])
        hol = self.get_holidays()
        T1 = T + datetime.timedelta(days=1)
        while((T1 in hol[exch]) or (T1.weekday() in [5, 6])):
            T1 = T1 + datetime.timedelta(days=1)
        return T1

    def tdate_shift(self, T, N, exch='sse'):
        '''
            func: 前寻或后寻第N个交易日
            T: 'yyyy-mm-dd' or dt.date
            N: 整数，负数表示向前（过去）寻找，正数表示向后（未来）寻找
            exch: 'sse', 'szse', 'nyse', 'lse', 'tse', 'hkex', 'fse'
            e.g. T1 = tdate_shift(dt.date(2020,3,11), 2, 'sse')
            out: dt.date
        '''

        # 检查输入参数
        if type(T) == str:
            T = datetime.datetime.strptime(T,'%Y-%m-%d').date()
        assert(type(N)==int)
        assert(exch in ['sse', 'szse', 'nyse', 'lse', 'tse', 'hkex', 'fse'])
        # 计算后一交易日
        T1 = T
        for n in range(abs(N)):
            T1 = self.prev_tdate(T1, exch) if N<0 else self.next_tdate(T1, exch)
        return T1


    def get_tdates(self, start_date=None, end_date=None, exch='sse'):
        '''
            func: 生成交易日序列
            start_date: 'yyyy-mm-dd' or dt.date
            end_date: 'yyyy-mm-dd' or dt.date
            exch: 'sse', 'szse', 'nyse', 'lse', 'tse', 'hkex', 'fse'
            e.g. dates = get_tdates('2020-1-1', '2020-3-1', 'sse')
            out: list of dt.date
            bug: 如果start_date太早，会因为没有holidays记录，而取出过多的交易日序列
        '''

        # 检查输入参数
        if start_date is None:
            start_date = datetime.date(1991,1,1)
        if end_date is None: #TODO 假期数据截止到2021年当年
            end_date = datetime.date(2022,1,1)
        start_date = pd.to_datetime(start_date).date()
        end_date = pd.to_datetime(end_date).date()
        if start_date > end_date:
            return []
        assert(exch in ['sse', 'szse', 'nyse', 'lse', 'tse', 'hkex', 'fse'])
        # 计算自然日序列
        dates = [start_date]
        while(dates[-1] < end_date):
            dates.append(dates[-1] + datetime.timedelta(days=1))
        hol = self.get_holidays()
        # 剔除节假日和周六日
        dates = [d for d in dates if ((d not in hol[exch]) and (d.weekday() not in [5, 6]))]
        return dates


if __name__ == '__main__':
    start_date = datetime.date(2019,1,1)
    end_date = datetime.date(2021,2,1)

    # 私募 信息 
    record_cd = ['SE1387','SE1383']
    #name_list = ['黑翼','CTA']
    #stg_name = ['股票策略']
    #stg_child = ['主观套利','指数增强']
    #company_name = ['黑翼']
    #start_date = datetime.date(2020,1,1)
    #mng_name = ['蒋彤','蒋锦志']
    # 选填
    ResearchApi().get_pf_info(record_cd=record_cd)

    # 私募净值
    record_cds = ['SE1387','SE1383']
    df = ResearchApi().get_pf_nav(record_cds=record_cds,
                                start_date=start_date,
                                end_date=end_date)
                            



import traceback
import pandas as pd
import datetime
# from .rq_raw_data_downloader import RqRawDataDownloader
from .em_raw_data_downloader import EmRawDataDownloader
from .web_raw_data_downloader import WebRawDataDownloader
from .oversea_data_downloader import OverseaDataUpdate
from .raw_data_helper import RawDataHelper
from .other_raw_data_downloader import OtherRawDataDownloader
from ...api.raw import RawDataApi


class RawDataDownloader:
    def __init__(self, rq_license=None, dry_run=False):
        self._data_helper = RawDataHelper()
        # self.rq_downloader = RqRawDataDownloader(rq_license, self._data_helper)
        self.web_downloader = WebRawDataDownloader(self._data_helper)
        self.em_downloader = EmRawDataDownloader(self._data_helper, dry_run)
        self.oversea_download = OverseaDataUpdate(update_period=10, data_per_page=5, data_helper=self._data_helper)
        self.oversea_download_long_term = OverseaDataUpdate(update_period=400, data_per_page=600, data_helper=self._data_helper)
        self.other_raw_data_downloader = OtherRawDataDownloader(data_helper=self._data_helper)

    def download(self, start_date, end_date):
        failed_tasks = []

        try:
            failed_tasks.extend(self.em_downloader.download_all(start_date, end_date))
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('unknown in em_downloader')

        # If 'em_tradedates' in failed_tasks, there is no trading day between start_date and end_date
        # Stop and return
        # try:
        #     failed_tasks.extend(self.rq_downloader.download_all(start_date, end_date))
        # except Exception as e:
        #     print(e)
        #     traceback.print_exc()
        #     failed_tasks.append('unknown in rq_downloader')
        try:
            failed_tasks.extend(self.oversea_download.update_fund_nav())
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('oversea fund nav')

        try:
            failed_tasks.extend(self.other_raw_data_downloader.hf_fund_nav_miss())
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('hf_fund_nav_miss')

        if 'em_tradedates' in failed_tasks:
            return failed_tasks
        else:

            try:
                failed_tasks.extend(self.web_downloader.download_all(start_date, end_date))
            except Exception as e:
                print(e)
                traceback.print_exc()
                failed_tasks.append('unknown in web_downloader')

            # 获取下一个交易日
            trading_day_df = RawDataApi().get_em_tradedates(start_date=end_date)
            if trading_day_df.shape[0] <= 1:
                print(f'get trading days start with {end_date} failed')
                failed_tasks.append('get_trading_day for weekly/monthly indicator in raw')
            else:
                next_trading_day = trading_day_df.iloc[1, :].TRADEDATES
                print(f'got next trading day {next_trading_day}')
                next_trading_dt = pd.to_datetime(next_trading_day, infer_datetime_format=True).date()
                end_date_dt = pd.to_datetime(end_date, infer_datetime_format=True).date()
                if end_date_dt.weekday() < next_trading_dt.weekday() and next_trading_dt < end_date_dt + datetime.timedelta(weeks=1):
                    # 表明本非周末，今天不需要更新
                    print(f'monthly indicator update on every weekend, not today {end_date_dt}')
                else:
                    if not self.oversea_download_long_term.update_fund_adj_factor():
                        failed_tasks.append('oversea_fund_adj_factor')
                    if not self.other_raw_data_downloader.fund_nav_adj_process():
                        failed_tasks.append('oversea_fund_nav_adj_process')
        return failed_tasks

    def get_updated_count(self):
        return self._data_helper._updated_count




from typing import Optional, Set, Union
# from functools import lru_cache
import pandas as pd
import time
from ...constant import FundFactorType, StockFactorType
from ...util.singleton import Singleton
from .calculator import Calculator
from .universe import Universe

_STOCK_BUCKET_NAME = 'tl-factors'
_FUND_BUCKET_NAME = 'tl-fund-factors'


class Factor(metaclass=Singleton):


    def __init__(self, f_name: str, f_type: Union[StockFactorType, FundFactorType], f_level: str = 'basic'):
        self._f_name: str = f_name
        self._s3_file_name: str = f'{f_name}.parquet'
        self._f_level = f_level
        if isinstance(f_type, StockFactorType):
            self._s3_uri: str = f's3://{_STOCK_BUCKET_NAME}/factor/'
            self._s3_uri_factor_ret: str = f's3://{_STOCK_BUCKET_NAME}/factor_ret/'
        elif isinstance(f_type, FundFactorType):
            self._s3_uri = f's3://{_FUND_BUCKET_NAME}/{self._f_level}/'
            self._s3_uri_factor_ret = None
        else:
            assert False, f'invalid factor type {f_type}'
        self._f_type: Union[StockFactorType, FundFactorType] = f_type
        self._factor: Optional[pd.DataFrame] = None
        self._deps: Set[Factor] = set()

    # @lru_cache
    def get(self, universe: str = 'default') -> Optional[pd.DataFrame]:
        # 这里我们认为没有必要去存储各个universe下的因子值
        # 因为只要全量(default)的因子值保存了，使用stock_list去过滤出相应universe下的因子值会非常快
        if self._factor is None:
            self._load_data()
        if universe != 'default':
            stock_list = Universe().get(universe)
            if stock_list is not None:
                return self._factor.loc[:, stock_list]
            print(f'get data of universe {universe} failed, use default universe')
        return self._factor

    # delete
    def clear(self, recursive=False):
        if recursive:
            for _dep in self._deps:
                _dep.clear(recursive)
        self._factor = None

    def get_normalized(self, universe: str = 'default'):
        data = self.get(universe)
        return ((data.T - data.mean(axis=1))/data.std(axis=1)).T

    @property
    def name(self):
        return self._f_name

    @property
    def f_type(self):
        return self._f_type

    def save(self) -> bool:
        if self._factor is None:
            return False
        self._factor.to_parquet(self._get_s3_factor_uri, compression='gzip')
        print(f'\t[upload] to s3 success {self._f_name} {self._get_s3_factor_uri}')
        return True

    def calc(self):
        assert False, f'should impl specialized calc algo for the {self._f_name} factor'

    def _load_data(self) -> Optional[pd.DataFrame]:
        try:
            t0 = time.time()
            self._factor: pd.DataFrame = pd.read_parquet(self._get_s3_factor_uri)
            t1 = time.time()
            print(f'\t[time][{self._f_level}] fetch factor: {self._f_name} from s3 cost {round(t1 - t0, 4)}')
        except Exception as e:
            print(f'retrieve data from s3 failed, (err_msg){e}; try to re-calc {self._f_name}')
            self.calc()
        return self._factor

    @property
    def _get_s3_factor_uri(self) -> str:
        return f'{self._s3_uri}{self._s3_file_name}'

    def _get_s3_factor_uri_with_universe(self, universe: str) -> str:
        return f'{self._s3_uri}with_universe/{universe}/{self._s3_file_name}'

    def _get_s3_factor_ret_uri(self, universe: str) -> str:
        return f'{self._s3_uri_factor_ret}{universe}/{self._s3_file_name}'



import datetime
import traceback
import numpy as np
import pandas as pd
import re
from multiprocessing import Pool
from .basic_data_helper import BasicDataHelper
from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...view.basic_models import FundIPOStats


class BasicIPOStat():
    '''打新基金'''
    def __init__(self, data_helper: BasicDataHelper):
        self._data_helper = data_helper
        self._raw_data_api = RawDataApi()
        self._basic_data_api = BasicDataApi()

    def init(self):
        self.fund_raw_ipo_status = self._raw_data_api.get_em_fund_ipo_status()
        self.fund_size = self._basic_data_api.get_fund_size_all_data()
        self.fund_info = self._basic_data_api.get_fund_info()
        self.desc_dic = self.fund_info[['fund_id', 'desc_name']].set_index('fund_id').to_dict()['desc_name']
        self.fund_info['trim_name'] = self.fund_info.desc_name.map(lambda x: re.subn(r'[ABCDEFR]{1,2}(\(人民币\)|\(美元现汇\)|\(美元现钞\)|1|2|3)?$', '', x)[0])
        self.fund_info_type = self.fund_info.set_index(['trim_name', 'desc_name'])

        self.fund_size = self.fund_size.pivot_table(index='datetime', columns='fund_id', values='size')
        d1 = self.fund_raw_ipo_status.end_date.unique().tolist()
        d2 = self.fund_size.index.unique().tolist()
        date_list = sorted(list(set(d1).union(d2)))
        self.fund_size = self.fund_size.reindex(date_list).fillna(method='ffill')
        self.fund_size = self.fund_size.stack().reset_index().rename(columns={'datetime': 'end_date', 0: 'size'}).set_index(['fund_id', 'end_date'])
        self.fund_raw_ipo_status['fund_id'] = self.fund_raw_ipo_status.apply(lambda x: self._data_helper._get_fund_id_from_order_book_id(x.em_id.split('.')[0], x.end_date), axis=1)
        self.fund_raw_ipo_status = self.fund_raw_ipo_status.drop(['em_id', '_update_time'], axis=1)
        self.fund_raw_ipo_status = self.fund_raw_ipo_status.dropna(subset=['fund_id'])
        self.fund_raw_ipo_status.loc[:, 'desc_name'] = self.fund_raw_ipo_status.fund_id.map(lambda x: self.desc_dic[x])

    def _loop_item(self, i):
        df_i = pd.DataFrame([i])  #.set_index(['fund_id'])
        desc_name = i['desc_name']
        index_1 = self.fund_info_type.loc[pd.IndexSlice[:, desc_name], :].index.get_level_values(0).values[0]
        fund_info_i = self.fund_info_type.loc[index_1]
        if fund_info_i.shape[0] == 1:
            try:
                df_i['size'] = self.fund_size.loc[i['fund_id'], pd.to_datetime(i['end_date'], infer_datetime_format=True)].values[0]
            except Exception:
                # print(f'[_loop_item] can not find fund size (info){df_i}')
                return
        else:
            df_i = df_i.set_index('fund_id')
            _fund_list = fund_info_i.fund_id.tolist()
            for _fund_id in _fund_list:
                df_i.loc[_fund_id] = df_i.loc[i['fund_id']]
            df_i = df_i.reset_index().set_index(['fund_id', 'end_date']).join(self.fund_size).dropna(subset=['size']).reset_index()
        df_i.loc[:, 'ipo_allocation_weight'] = 10000 * df_i['ipo_allocation_amount'] / df_i['size'].sum()
        return df_i

    def calculate_history(self, end_date: str = ''):
        if end_date:
            self.fund_raw_ipo_status = self.fund_raw_ipo_status[self.fund_raw_ipo_status.end_date == pd.to_datetime(end_date, infer_datetime_format=True).date()]
        dict_list = self.fund_raw_ipo_status.to_dict('records')
        p = Pool()
        res = [i for i in p.imap_unordered(self._loop_item, dict_list, 2000) if i is not None]
        p.close()
        p.join()
        self.result = pd.concat(res).drop('desc_name', axis=1)
        res = []
        for i in self.result.end_date:
            if isinstance(i, pd.Timestamp):
                res.append(i.date())
            else:
                res.append(i)
        self.result['end_date'] = res
        self.result = self.result.drop_duplicates(subset=['fund_id', 'end_date'])
        self._data_helper._upload_basic(self.result, FundIPOStats.__table__.name)


if __name__ == "__main__":
    bipos = BasicIPOStat(BasicDataHelper())
    bipos.init()
    bipos.calculate_history('20210121')




import datetime
from typing import List
import traceback
import pandas as pd

from .fund_indicator_processor import FundIndicatorProcessor
# from .derived_index_val import IndexValProcess
from .derived_index_val_long_term import IndexValProcessLongTerm
from .style_analysis_processor import StyleAnalysisProcessor
from .derived_data_helper import DerivedDataHelper
from .fund_indicator_processor_weekly import FundIndicatorProcessorWeekly
from .fund_indicator_processor_monthly import FundIndicatorProcessorMonthly
from .derived_indicators_processor_group import FundIndicatorProcessorGroup
from .derived_index_ret_vol import IndexIndicatorProcessor
from .derived_fund_aic import AIPProcess
from .derived_mng_fund_rank import FundManagerFundRankProcessor
from .derived_fund_alpha import FundAlphaProcessor
# from .derived_fund_score_extended import FundScoreExtendedProcess
from .derived_fund_score_group import FundScoreGroupProcess
from .derived_fund_score_group_v1 import FundScoreGroupV1Process
from .barra_cne5_factor_processor import BarraCNE5FactorProcessor
from .style_reg_processer import StyleBoxReg
from .style_box import StyleBoxGenerator
from .fund_manager_processor import ManagerProcessor
from .derived_top_manager_funds import TopManagerFundProcessor
from .derived_fund_industry_exposure import FundIndustryExposureProcess
from .third_party_portfolio.qieman import QiemanPortfolioProcessor
from .third_party_portfolio.danjuan import DanjuanPortfolioProcessor
from .third_party_portfolio.em import EmPortfolioProcessor
from .third_party_portfolio.gf_sec import GFSecPortfolioProcessor
from .third_party_portfolio.manually_crawled import ManuallyCrawledPortfolioProcessor
from .fund_manager_processor_advanced import ManagerProcessorDev
from .derived_oversea_nav_analysis import OverseaFundNavAnalysis
from. derived_future_main_chain_price import FutureMainForcePrice
from ...api.basic import BasicDataApi
from ..raw.raw_data_helper import RawDataHelper


class DerivedDataProcessor:
    def __init__(self):
        self._data_helper = DerivedDataHelper()
        self._raw_data_helper = RawDataHelper()
        self.fund_indicator_processor = FundIndicatorProcessor(self._data_helper)
        self.fund_indicator_processor_weekly = FundIndicatorProcessorWeekly(self._data_helper)
        self.fund_indicator_processor_monthly = FundIndicatorProcessorMonthly(self._data_helper)
        self.fund_indicator_processor_group_1Y = FundIndicatorProcessorGroup(self._data_helper, 1)
        self.fund_indicator_processor_group_5Y = FundIndicatorProcessorGroup(self._data_helper, 5)
        self._manager_processor = ManagerProcessor(self._data_helper)
        # self.index_val_processor = IndexValProcess(self._data_helper)
        self.index_val_processor_long_term = IndexValProcessLongTerm(self._data_helper)
        self.index_indicator = IndexIndicatorProcessor(self._data_helper)
        self.fund_aip = AIPProcess(self._data_helper)
        # 这版基金评分先不用了，这里先注释掉
        # self._derived_fund_score_extended = FundScoreExtendedProcess(self._data_helper)
        self._derived_fund_score_group_v1_1Y = FundScoreGroupV1Process(self._data_helper, 1)
        self._derived_fund_score_group_v1_5Y = FundScoreGroupV1Process(self._data_helper, 5)
        self._derived_fund_score_group_1Y = FundScoreGroupProcess(self._data_helper, 1)
        self._derived_fund_score_group_5Y = FundScoreGroupProcess(self._data_helper, 5)
        self._barra_cne5_factor_processor = BarraCNE5FactorProcessor(self._data_helper)
        self._style_regression_processor = StyleBoxReg(self._data_helper)
        self._style_box_processor = StyleBoxGenerator(self._data_helper)
        self.style_analysis_processors: List[StyleAnalysisProcessor] = []
        self.mng_fund_rank = FundManagerFundRankProcessor(self._data_helper)
        self._fund_alpha_processor = FundAlphaProcessor(self._data_helper)
        self.top_manager_funds_processor = TopManagerFundProcessor(self._data_helper)
        self._fund_industry_exposure = FundIndustryExposureProcess(self._data_helper)
        # 暂时只算这三个universe
        for universe in ('hs300', 'csi800', 'all'):
            sap = StyleAnalysisProcessor(self._data_helper, universe)
            self.style_analysis_processors.append(sap)
        self._qieman_portfolio_processor = QiemanPortfolioProcessor(self._data_helper)
        self._danjuan_portfolio_processor = DanjuanPortfolioProcessor(self._data_helper)
        self._em_portfolio_processor = EmPortfolioProcessor(self._data_helper)
        self._gf_sec_portfolio_processor = GFSecPortfolioProcessor(self._data_helper)
        self._manually_crawled_portfolio_processor = ManuallyCrawledPortfolioProcessor(self._data_helper)
        self._manager_processor_dev = ManagerProcessorDev(self._data_helper)
        self._oversea_fund_nav_analysis = OverseaFundNavAnalysis(data_helper=self._raw_data_helper)
        self._future_main_chain_price = FutureMainForcePrice(self._data_helper)

    def process_all(self, start_date, end_date):
        failed_tasks = []
        try:
            failed_tasks.extend(self.fund_indicator_processor.process(start_date, end_date))
            # failed_tasks.extend(self.index_val_processor.process(start_date, end_date))
            failed_tasks.extend(self.index_val_processor_long_term.process(start_date, end_date))
            failed_tasks.extend(self.index_indicator.process(end_date))
            failed_tasks.extend(self._manager_processor.process(end_date))
            failed_tasks.extend(self.fund_aip.process(start_date, end_date))
            failed_tasks.extend(self._barra_cne5_factor_processor.process(start_date, end_date))
            failed_tasks.extend(self._style_regression_processor.process(start_date, end_date))
            failed_tasks.extend(self._style_box_processor.process(start_date, end_date))
            failed_tasks.extend(self.mng_fund_rank.process(end_date))
            failed_tasks.extend(self._fund_alpha_processor.process(start_date, end_date))
            failed_tasks.extend(self.top_manager_funds_processor.process(start_date, end_date))
            for sap in self.style_analysis_processors:
                failed_tasks.extend(sap.process(start_date, end_date))
            failed_tasks.extend(self._qieman_portfolio_processor.process(end_date))
            failed_tasks.extend(self._danjuan_portfolio_processor.process(end_date))
            failed_tasks.extend(self._em_portfolio_processor.process(end_date))
            failed_tasks.extend(self._gf_sec_portfolio_processor.process(end_date))
            failed_tasks.extend(self._manually_crawled_portfolio_processor.process(end_date))
            failed_tasks.extend(self._oversea_fund_nav_analysis.daily_process(end_date))

            # 获取下一个交易日
            api = BasicDataApi()
            trading_day_df = api.get_trading_day_list(start_date=end_date)
            if trading_day_df.shape[0] <= 1:
                print(f'get trading days start with {end_date} failed')
                failed_tasks.append('get_trading_day for weekly/monthly indicator in derived')
            else:
                next_trading_day = trading_day_df.iloc[1, :].datetime
                print(f'got next trading day {next_trading_day}')
                end_date_dt = pd.to_datetime(end_date, infer_datetime_format=True).date()
                next_trading_dt = pd.to_datetime(next_trading_day, infer_datetime_format=True).date()
                if end_date_dt.weekday() < next_trading_dt.weekday() and next_trading_dt < end_date_dt + datetime.timedelta(weeks=1):
                    # 表明本周后边还有交易日，今天不需要更新
                    print(f'weekly indicator only update on the last day of week, not today {end_date_dt}')
                else:
                    failed_tasks.extend(self.fund_indicator_processor_weekly.process(start_date, end_date, end_date_dt))
                    # 分别用2年和5年的数据算一下
                    failed_tasks.extend(self.fund_indicator_processor_group_1Y.process(start_date, end_date, end_date_dt))
                    failed_tasks.extend(self.fund_indicator_processor_group_5Y.process(start_date, end_date, end_date_dt))
                    # failed_tasks.extend(self._derived_fund_score_extended.process(start_date, end_date, end_date_dt))
                    failed_tasks.extend(self._derived_fund_score_group_v1_1Y.process(start_date, end_date, end_date_dt))
                    failed_tasks.extend(self._derived_fund_score_group_v1_5Y.process(start_date, end_date, end_date_dt))
                    failed_tasks.extend(self._derived_fund_score_group_1Y.process(start_date, end_date, end_date_dt))
                    failed_tasks.extend(self._derived_fund_score_group_5Y.process(start_date, end_date, end_date_dt))
                    failed_tasks.extend(self._fund_industry_exposure.process(start_date, end_date, end_date_dt))
                    failed_tasks.extend(self._oversea_fund_nav_analysis.weekly_process(end_date))

                if end_date_dt.year == next_trading_dt.year and end_date_dt.month == next_trading_dt.month:
                    # 表明本月后边还有交易日，今天不需要更新
                    print(f'monthly indicator only update on the last day of month, not today {end_date_dt}')
                else:
                    failed_tasks.extend(self.fund_indicator_processor_monthly.process(start_date, end_date, end_date_dt))
                    failed_tasks.extend(self._manager_processor_dev.process(end_date))

        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('unknown in derived_data_processor')
        return failed_tasks

    def process_not_trading_day(self, start_date, end_date):
        failed_tasks = []
        try:
            failed_tasks.extend(self.index_val_processor_long_term.process(start_date, end_date))
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('unknown in derived_data_processor')
        return failed_tasks

    def get_updated_count(self):
        _dic = self._raw_data_helper._updated_count
        for k, v in _dic.items():
            self._data_helper._updated_count[k] = v
        return self._data_helper._updated_count


if __name__ == '__main__':
    ddp = DerivedDataProcessor()
    start_date = '20200820'
    end_date = '20200820'
    ddp.process_all(start_date, end_date)
    # ddp.fund_indicator_processor.process(start_date, end_date)
    # ddp.fund_score_processor.process(start_date, end_date)
    # import pandas as pd
    # date_list = pd.date_range(end='2010-05-31', periods=65, freq='M').sort_values(ascending=False).to_pydatetime()
    # for date in date_list:
    #     date = date.strftime('%Y%m%d')
    #     ddp.fund_indicator_processor_monthly.process(date, date)




from surfing.structure import Tick
import pandas as pd
import numpy as np
import os
import json
import datetime

class DataWrapper:

    def __init__(self, base_path):
        assert base_path is not None, 'data path has to be set'
        self.base_path = base_path
        self.meta_info = DataWrapper.load_meta(base_path)

    @staticmethod
    def load_meta(folder):
        file_path = folder + '/.meta_data.json'
        try:
            with open(file_path,'r') as load_f:
                meta_info = json.load(load_f)
        except:
            assert 0, 'meta info not exsited'
        return meta_info

    @staticmethod
    def save_data(df, folder, ticker, colomns=None):
        df['time'] = np.array(df.index.to_pydatetime(), dtype='datetime64[ms]')
        data_type = [(col_i, 'f4') for col_i in colomns]
        data_type.append(('time', 'datetime64[ms]'))
        df = [tuple(x) for x in df.values]
        df = np.array(df, dtype=data_type)
        np.save(os.path.join(folder, '{}.npy'.format(ticker)), df)

    @staticmethod
    def has_data(folder, ticker):
        return os.path.isfile('{}/{}.npy'.format(folder, ticker))

    @staticmethod
    def base_load_data(folder, ticker, index_column, colomns=None):
        # may learn to cache data here!
        if not DataWrapper.has_data(folder, ticker):
            return None
        value = np.load(os.path.join(folder, '{}.npy'.format(ticker)), allow_pickle=True)
        df = pd.DataFrame(data=value, columns=colomns)
        df.set_index(index_column, drop=True, inplace=True)
        df.fillna(method='ffill', inplace=True)
        return df

    @staticmethod
    def base_load_npz(folder, price, columns):
        file_i = f'{folder}/{price}.npz'
        data_i_folder = os.path.isfile(file_i)
        if not data_i_folder:
            return None
        value = np.load(file_i, allow_pickle=True)
        df = pd.DataFrame(data=value['arr_0'], columns=columns)
        return df
